from plotly import tools as ptools
import plotly.io as pio

import plotly.graph_objs as go

import esppy.espapi.connections as connections
import esppy.espapi.tools as tools
import esppy.espapi.viewers as viewers

from esppy.espapi.tools import Options, Colors

import sys

import threading

import datetime

import base64

import ipywidgets as widgets
import ipyleaflet as maps

import numpy as np

import logging

import re

import math

import random

#['ggplot2', 'seaborn', 'plotly', 'plotly_white', 'plotly_dark', 'presentation', 'xgridoff', 'none']

pio.templates.default = "none"

class Visuals(Options):

    _dataHeader = "_data://"

    def __init__(self,**kwargs):
        Options.__init__(self,**kwargs)
        self._visuals = []
        self._border = None

        if self.hasOpt("colormap"):
            self._colors = tools.Colors(colormap=self.getOpt("colormap"))
        elif self.hasOpt("colors"):
            self._colors = tools.Colors(colors=self.getOpt("colors"))
        else:
            self._colors = tools.Colors(colormap=None)

        self._titleStyle = tools.Options(font_size="12pt")

        self._axisWidth = 1

    def optionSet(self,name,value):
        if name == "colormap":
            colormap = value
            self._colors = tools.Colors(colormap=colormap)

    def createBarChart(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = BarChart(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createLineChart(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = LineChart(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createTimeSeries(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = TimeSeries(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createBubbleChart(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = BubbleChart(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createPieChart(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = PieChart(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createMap(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = Map(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createGauge(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = Gauge(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createCompass(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = Compass(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createTable(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = Table(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createImages(self,datasource,**kwargs):
        datasource.addDelegate(self)
        chart = Images(self,datasource,**kwargs)
        chart.create()
        self._visuals.append(chart)
        return(chart)

    def createControls(self,datasource,**kwargs):
        datasource.addDelegate(self)
        controls = Controls(self,datasource,**kwargs)
        controls.create()
        self._visuals.append(controls)
        return(controls)

    def createModelViewer(self,connection,**kwargs):
        return(viewers.ModelViewer(self,connection,**kwargs))

    def createLogViewer(self,connection,**kwargs):
        return(viewers.LogViewer(self,connection,**kwargs))

    def createStatsViewer(self,connection,**kwargs):
        return(viewers.StatsViewer(self,connection,**kwargs))

    def dataChanged(self,datasource,data,clear):
        for v in self._visuals:
            if v._datasource == datasource:
                v.draw(data,clear)

    def infoChanged(self,datasource):
        for v in self._visuals:
            if v._datasource == datasource:
                v.info()

    def handleStats(self,datasource):
        for v in self._visuals:
            if v._datasource == datasource:
                v.draw()

    def clear(self):
        self._visuals = []

    def formatTitle(self,text):
        content = ""
        content += "<div style='"
        i = 0
        for k,v in self._titleStyle.options.items():
            s = k.replace("_","-")
            if i > 0:
                content += ";"
            content += s + ":" + v
            i += 1

        content += ";text-align:center"
        content += "'>"
        content += text
        content += "</div>"

        return(content)

class Chart(Options,widgets.VBox):
    def __init__(self,visuals,datasource,**kwargs):
        Options.__init__(self,**kwargs)
        widgets.VBox.__init__(self,layout=widgets.Layout(width=self.getOpt("width","95%"),height=self.getOpt("height","95%"),border=visuals.getOpt("border","0"),padding=visuals.getOpt("padding","10px"),margin=visuals.getOpt("margin","10px")))
        self._visuals = visuals
        self._datasource = datasource
        self._figure = None
        self._data = None
        self._layout = None
        self._controls = None
        self._title = widgets.HTML(layout=widgets.Layout(overflow="hidden"))
        self._banner = widgets.VBox([self._title])

    def setOpts(self,**kwargs):
        Options.setOpts(self,**kwargs)
        self.create()

    def setWidth(self,value):
        self.setOpt("width",value)

    def setHeight(self,value):
        self.setOpt("height",value)

    def createContent(self):
        pass

    def create(self):
        self.createContent()

        if self._data != None:
            self._layout = go.Layout()

            margin = 20
            self._layout["margin"] = dict(l=margin,r=margin,b=margin,t=margin)

            xRange = self.getOpt("xrange")
            if xRange != None:
                self._layout["xaxis"]["range"] = xRange
            yRange = self.getOpt("yrange")
            if yRange != None:
                self._layout["yaxis"]["range"] = yRange

            self._layout["xaxis"]["showticklabels"] = self.getOpt("showticks",True)
            self._layout["xaxis"]["showline"] = False

            self._figure = go.FigureWidget(data=self._data,layout=self._layout)

            children = [self._banner,self._figure]

            if self.getOpt("show_controls",False):
                if self._controls == None:
                    self._controls = ControlPanel(self._datasource) 
                children.append(self._controls)

            self.children = children

        self.draw(None,True)

    def setTitle(self,title = None):

        if title == None:
            title = self.getOpt("title")

        if title == None:
            title = self._datasource._path

        if isinstance(self._datasource,connections.EventCollection):
            if self._datasource._pages > 1:
                    title += " (Page " + str(self._datasource._page + 1) + " of " + str(self._datasource._pages) + ")"

        filter = self._datasource.getOpt("filter")

        if filter != None:
            title += "<br>"
            title += filter

        self._title.value = self._visuals.formatTitle(title)

    def createMarkers(self):
        o = {}
        marker = {}

        keys = self._datasource.getKeyValues()

        text = []

        for i,key in enumerate(keys):
            text.append(key)

        size = None

        value = self.getOpt("size")

        if value != None: 
            try:
                num = int(value)
                marker["size"] = num
            except:
                size = value

        color = self.getOpt("color")

        if size != None or color != None:

            if size != None:
                s = self._datasource.getValues(size)
                if s != None and len(s) > 0:
                    maxsize = 60.
                    minsize = 5
                    marker["size"] = s
                    marker["sizemode"] = "area"
                    marker["sizeref"] = 2. * max(s) / (maxsize ** 2)
                    marker["sizemin"] = minsize

                    for i,v in enumerate(s):
                        text[i] += "<br>"
                        text[i] += size + "=" + str(v)

            color = self.getOpt("color")

            if color != None:
                s = self._datasource.getValues(color)
                if s != None and len(s) > 0:
                    marker["color"] = s
                    marker["showscale"] = True
                    marker["colorscale"] = self._visuals._colors.colorscale

                    if size == None or color != size:
                        for i,v in enumerate(s):
                            if size != None and v in size == False:
                                text[i] += "<br>"
                            text[i] += color + "=" + str(v)

        #marker["line"] = {"width":2,"color":"#ff0000"}

        o["marker"] = marker
        o["text"] = text

        return(o)

    def info(self):
        if self._controls != None:
            self._controls.processInfo()
        self.setTitle()

    def getValues(self,name):
        values = []

        value = self.getOpt(name)

        if value != None:
            if type(value) is list:
                for v in value:
                    values.append(v)
            else:
                values.append(value)

        return(values)

class BarChart(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)

    def createContent(self):
        values = self.getValues("y")
        colors = self._visuals._colors.getFirst(len(values))
        opacity = self.getOpt("opacity")

        self._data = []

        orientation = self.getOpt("orientation","vertical")

        if orientation == "horizontal":
            for i,v in enumerate(values):
                self._data.append(go.Bar(x=[0],y=[""],name=v,orientation="h",marker_color=colors[i]))

        else:
            for i,v in enumerate(values):
                self._data.append(go.Bar(x=[""],y=[0],name=v,opacity=opacity,marker_color=colors[i]))

    def draw(self,data = None,clear = False):
        if self._figure == None:
            return

        self._figure.update_xaxes(showline=True,linewidth=self._visuals._axisWidth)
        self._figure.update_yaxes(showline=True,linewidth=self._visuals._axisWidth)

        x = self.getValues("x")
        values = self.getValues("y")

        orientation = self.getOpt("orientation","vertical")

        if orientation == "horizontal":

            if len(x) > 0:
                try:
                    data = self._datasource.getValuesBy(x,values)
                except:
                    return

                for i,v in enumerate(values):
                    self._figure.data[i].x = data["values"][v]
                    self._figure.data[i].y = data["keys"]

            else:
                keys = self._datasource.getKeyValues()

                for i,v in enumerate(values):
                    y = self._datasource.getValues(v)
                    self._figure.data[i].x = y
                    self._figure.data[i].y = keys

        else:

            if len(x) > 0:
                try:
                    data = self._datasource.getValuesBy(x,values)
                except:
                    return

                for i,v in enumerate(values):
                    self._figure.data[i].x = data["keys"]
                    self._figure.data[i].y = data["values"][v]
            else:
                keys = self._datasource.getKeyValues()

                if len(keys) == 0:
                    for i,v in enumerate(values):
                        self._figure.data[i].x = [""]
                        self._figure.data[i].y = [0]
                else:
                    for i,v in enumerate(values):
                        y = self._datasource.getValues(v)
                        self._figure.data[i].x = keys
                        self._figure.data[i].y = y

        self._figure.update_xaxes(automargin=True)
        self._figure.update_yaxes(automargin=True)

        self.setTitle()

class LineChart(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)

    def createContent(self):
        values = self.getValues("y")

        self._data = []

        width = self.getOpt("line_width",2)
        shape = "linear"
        if self.getOpt("curved",False):
            shape = "spline"
        line = {"width":width,"shape":shape}

        fill = self.getOpt("fill",False)

        colors = self._visuals._colors.getFirst(len(values))

        for i,v in enumerate(values):
            if fill:
                if i == 0:
                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="none",fill="tozeroy",fillcolor=colors[i]))
                else:
                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="none",fill="tonexty",fillcolor=colors[i]))
            else:
                line["color"] = colors[i]
                self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="lines",line=line))

    def draw(self,data = None,clear = False):
        if self._figure == None:
            return

        self._figure.update_xaxes(showline=True,linewidth=self._visuals._axisWidth)
        self._figure.update_yaxes(showline=True,linewidth=self._visuals._axisWidth)

        values = self.getValues("y")
        x = self.getValues("x")

        if len(x) > 0:
            try:
                data = self._datasource.getValuesBy(x,values)
            except Exception as e:
                logging.info(str(e))
                return

            for i,v in enumerate(values):
                self._figure.data[i].x = data["keys"]
                self._figure.data[i].y = data["values"][v]
        else:
            keys = self._datasource.getKeyValues()

            for i,v in enumerate(values):
                y = self._datasource.getValues(v)
                self._figure.data[i].x = keys
                self._figure.data[i].y = y

        self._figure.update_xaxes(automargin=True)
        self._figure.update_yaxes(automargin=True)

        self.setTitle()

class TimeSeries(LineChart):
    def __init__(self,visuals,datasource,**kwargs):
        LineChart.__init__(self,visuals,datasource,**kwargs)

        if self.hasOpt("time") == False:
            raise Exception("must specify time field for a TimeSeries")
 
        self.setOpt("x",self.getOpt("time"))

class PieChart(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)

    def createContent(self):
        value = self.getValues("value")

        self._data = []

        if len(value) == 1:
            self._data.append(go.Pie(labels=[""],values=[0],name=value[0]))

    def draw(self,data = None,clear = False):
        if self._figure == None:
            return

        labels = self.getValues("labels")
        value = self.getValues("value")

        if len(value) == 1:
            if len(labels) > 0:
                data = self._datasource.getValuesBy(labels,value)

                self._figure.data[0].labels = data["keys"]
                self._figure.data[0].values = data["values"][value[0]]

            else:
                keys = self._datasource.getKeyValues()

                v = self._datasource.getValues(value[0])
                self._figure.data[0].labels = keys
                self._figure.data[0].values = v

        self.setTitle()

class BubbleChart(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)

    def createContent(self):
        values = self.getValues("y")

        self._data = []

        for v in values:
            self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="markers"))

    def draw(self,data = None,clear = False):
        if self._figure == None:
            return

        self._figure.update_xaxes(showline=True,linewidth=self._visuals._axisWidth)
        self._figure.update_yaxes(showline=True,linewidth=self._visuals._axisWidth)

        data = None

        x = self.getValues("x")
        if len(x) == 0:
            x = self._datasource.getKeyFieldNames()
        values = self.getValues("y")
        size = self.getOpt("size")
        color = self.getOpt("color")

        a = []
        a.extend(values)

        if size != None:
            if (size in a) == False:
                a.append(size)

        if color != None:
            if (color in a) == False:
                a.append(color)

        try:
            data = self._datasource.getValuesBy(x,a)
        except:
            return

        keys = data["keys"]

        if len(keys) == 0:
            for i,v in enumerate(values):
                self._figure.data[i].x = [""]
                self._figure.data[i].y = [0]
                return

        marker = {}

        text = None

        if size != None or color != None:

            text = []

            for i in range(0,len(keys)):
                text.append("")

            if size != None:
                if size in data["values"]:
                    s = data["values"][size]
                    if s != None and len(s) > 0:
                        maxsize = 60.
                        minsize = 5
                        marker["size"] = s
                        marker["sizemode"] = "area"
                        marker["sizeref"] = 2. * max(s) / (maxsize ** 2)
                        marker["sizemin"] = minsize

                        for i,v in enumerate(s):
                            text[i] += size + "=" + str(v)

            color = self.getOpt("color")

            if color != None:
                if color in data["values"]:
                    s = data["values"][color]
                    if s != None:
                        marker["color"] = s
                        marker["showscale"] = True
                        marker["colorscale"] = self._visuals._colors.colorscale

                        for i,v in enumerate(s):
                            if size != None:
                                text[i] += "<br>"
                            text[i] += color + "=" + str(v)

        for i,v in enumerate(values):
            self._figure.data[i].x = keys
            self._figure.data[i].y = data["values"][v]
            self._figure.data[i].marker = marker
            self._figure.data[i].text = text

        self._figure.update_xaxes(automargin=True)
        self._figure.update_yaxes(automargin=True)

        self.setTitle()

class Table(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)
        self._header = widgets.HTML(layout=widgets.Layout(overflow="hidden",margin="0",padding="0"))
        self._table = widgets.HTML(layout=widgets.Layout(overflow="auto",height="90%",margin="0",padding="0"))
        self.children = [self._banner,self._header,self._table]

    def draw(self,data = None,clear = False):
        allfields = self._datasource.getFields()
        columns = self.getValues("values")

        fields = []

        if len(columns) > 0:
            for column in columns:
                f = self._datasource.getField(column)
                if f != None:
                    fields.append(f)
        else:
            for f in allfields:
                fields.append(f)

        if len(fields) == 0:
            return

        width = int(100 / len(fields))

        border = "1px solid #d8d8d8"

        content = ""
        content += "<table cellspacing='0' cellpadding='4' style='border:" + border + ";width:100%'>"
        content += "<tr>"

        for f in fields:
            content += "<td style='border:" + border
            content += ";border-bottom:0"
            if f["isNumber"]:
                content += ";text-align:right"
            content += "' width='" + str(width) + "%'>" + f["name"] + "</td>"
            content += "\n"

        content += "</tr>"
        content += "</table>"
        self._header.value = content

        items = self._datasource.getList()

        content = ""
        content += "<table cellspacing='0' cellpadding='4' style='border:" + border + ";width:100%'>"
        #content += "<tr>"

        gradient = None

        color = self.getOpt("color")

        if color != None:
            if len(items) > 0:
                a = []

                for o in items:
                    if color in o:
                        a.append(float(o[color]))
                    else:
                        a.append(0)

                baseColor = self.getOpt("start_color")
                if baseColor == None:
                    baseColor = self._visuals._colors.lightest
                gradient = tools.Gradient(baseColor,levels=100,min=min(a),max=max(a))

        if self.getOpt("reversed",False):
            l = reversed(items)
        else:
            l = items

        for i,o in enumerate(l):
            content += "<tr"
            if color != None and gradient != None:
                if color in o:
                    content += " style='"
                    value = o[color]
                    c = gradient.darken(float(value))
                    content += "background:" + c
                    luma = Colors.getLuma(c)
                    if luma < 170:
                        content += ";color:white"
                    else:
                        content += ";color:black"
                    content += "'"
            content += ">"
            for f in fields:
                name = f["name"]
                value = o[name]
                if f["type"] == "blob":
                    if value.find(Visuals._dataHeader) == 0:
                        s = value[len(Visuals._dataHeader):]
                        index = s.find(":")
                        format = s[0:index]
                        s = s[index + 1:]
                        #value = "<div style='width:" + str(self.getOpt("image_width",400)) + "px;height:" + str(self.getOpt("image_height",400)) + "px;position:relative;margin:auto'>"
                        #value = "<div style='width:" + str(self.getOpt("image_width","80%")) + ";height:" + str(self.getOpt("image_height",400)) + "px;position:relative;margin:auto'>"
                        value = "<div style='width:" + str(width) + "%;position:relative;margin:auto'>"
                        value += "<img style='width:100%;height:100%' src='data:image/" + format + ";base64," + s + "'/>"
                        if "_nObjects_" in o:
                            numObjects = int(float(o["_nObjects_"]))
                            for j in range(0,numObjects):
                                s = "_Object" + str(j) + "_"
                                text = o[s].strip()
                                s = "_Object" + str(j) + "_x"
                                x = int(float(o[s]) * 100)
                                s = "_Object" + str(j) + "_y"
                                y = int(float(o[s]) * 100)
                                div = "<div style='position:absolute;left:" + str(x) + "%;top:" + str(y) + "%;"
                                div += "color:" + self.getOpt("image_text_color","black") + ";"
                                div += "'>"
                                div += text
                                div += "</div>"
                                value += div

                        value += "</div>"
                elif f["isTime"]:
                    num = int((int)(value) / 1000000)
                    date = datetime.datetime.fromtimestamp(num)
                    value = str(date)
                content += "<td style='border:" + border
                if f["isNumber"]:
                    content += ";text-align:right"
                if i == 0:
                    content += ";border-top:0"
                content += "' width='" + str(width) + "%'>" + str(value) + "</td>"
            content += "</tr>"
            content += "\n"

        content += "</table>"
        self._table.value = content

        self.setTitle()

class Images(Chart):

    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)
        self.layout.overflow = "auto"
        if self.hasOpt("image") == False:
            raise Exception("you must specify the image property")
        self._entries = {}
        self._lock = threading.Lock()
        self._imageWidth = self.getOpt("image_width",300)
        self._imageHeight = self.getOpt("image_height",300)
        orientation = self.getOpt("orientation","vertical")

        if orientation == "horizontal":
            self.layout.width = self.getOpt("width","800px")
            self.layout.height = str(self._imageHeight + 100) + "px"
        else:
            self.layout.width = str(self._imageWidth + 80) + "px"
            self.layout.height = self.getOpt("height","800px")

        self._detection = None

    def createContent(self):
        self.children = [self._banner]

    def draw(self,data = None,clear = False):
        if self._detection == None:
            if self._datasource.schema.hasFields():
                self._detection = (self._datasource.schema.getField("_nObjects_") != None)

        field = self.getOpt("image")

        data = self._datasource.getList()

        layout = False

        if clear:
            self._entries = []
            layout = True

        keys = []

        for o in data:
            if ("_key_" in o) == False:
                continue

            key = o["_key_"]

            if "_opcode" in o and o["_opcode"] == "delete":
                if self.removeEntry(key) != None:
                    layout = True
            else:
                keys.append(key)

                entry = self.getEntry(key)

                if entry == None:
                    entry = ImageEntry(self)
                    entry.data = o
                    entry.setOpt("title",key)
                    self._entries.append(entry)
                    layout = True

        remove = []

        for entry in self._entries:
            if (entry.key in keys) == False:
                remove.append(entry.key)
                layout = True

        for key in remove:
            self.removeEntry(key)

        if layout:
            orientation = self.getOpt("orientation","vertical")
            html = ""
            html += "<div style='overflow:visible'>"
            html += "<table cellspacing='10' cellpadding='10'>"
            if orientation == "horizontal":
                html += "<tr>"

            for entry in reversed(self._entries):
                if orientation == "vertical":
                    html += "<tr>"
                html += "<td>"
                html += entry.html
                html += "</td>"
                if orientation == "vertical":
                    html += "</tr>"
            if orientation == "horizontal":
                html += "</tr>"
            html += "</table>"
            html += "</div>"

            images = widgets.HTML(value=html,layout=widgets.Layout(border=self._visuals.getOpt("border","1px solid #d8d8d8"),width="100%",height="100%",overflow="auto",))

            self.children = [self._banner,images]

        self.layout.overflow = "auto"
        self.setTitle()

    def getEntry(self,key):
        for entry in self._entries:
            if entry.key == key:
                return(entry)

        return(None)

    def removeEntry(self,key):
        for i,entry in enumerate(self._entries):
            if entry.key == key:
                del self._entries[i]
                return(entry)

        return(None)

class ImageEntry(Options):
    def __init__(self,images,**kwargs):
        Options.__init__(self,**kwargs)
        self._images = images
        self._data = None
        self._key = ""
        self._title = widgets.HTML()
        self._html = None

    @property
    def html(self):
        if self._html == None:
            html = ""

            if self._data == None:
                return(html)

            field = self._images.getOpt("image")

            if field in self._data:
                data = self._data[field]

                if data.find(Visuals._dataHeader) == 0:
                    s = data[len(Visuals._dataHeader):]
                    key = self._data["_key_"]
                    index = s.find(":")
                    format = s[0:index]
                    s = s[index + 1:]

                    html += "<div style='width:" + str(self._images.getOpt("image_width",400)) + "px;height:" + str(self._images.getOpt("image_height",400)) + "px;position:relative;margin:auto;border:" + self._images.getOpt("image_border","1px solid #000000") + "'>"
                    html += "<img style='width:100%;height:100%' src='data:image/" + format + ";base64," + s + "'/>"

                    if self._images._detection:
                    #if False:
                        if "_nObjects_" in self._data:
                            value = str(self._data["_nObjects_"])
                            numObjects = int(float(value))
                            for i in range(0,numObjects):
                                s = "_Object" + str(i) + "_"
                                text = self._data[s].strip()
                                s = "_Object" + str(i) + "_x"
                                x = int(float(self._data[s]) * 100)
                                s = "_Object" + str(i) + "_y"
                                y = int(float(self._data[s]) * 100)
                                html += "<div style='position:absolute;zindex:1000;font-weight:normal;color:" + self._images.getOpt("label_color","white") + ";left:" + str(x) + "%;top:" + str(y) + "%;'>" + text + "</div>"

                    html += "</div>"

            self._html = html

        #logging.info(self._html)

        return(self._html)

    @property
    def key(self):
        return(self._key)

    @property
    def data(self):
        return(self._data)
    
    @data.setter
    def data(self,value):
        self._data = value
        if "_key_" in self._data:
            self._key = self._data["_key_"]

class Map(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)

        self._map = maps.Map()
        self._lat = None
        self._lon = None
        self._markers = {}
        self._colorbar = None

        self._colors = None
        self._colorRange = None

        self._circles = []
        self._polygons = []

        self._icon = None

    def createContent(self):
        if self.hasOpt("center"):
            self._map.center = self.getOpt("center")
        if self.hasOpt("zoom"):
            self._map.zoom = self.getOpt("zoom")

        self._lat = self.getOpt("lat")
        if self._lat == None:
            raise Exception("you must specify the lat property")

        self._lon = self.getOpt("lon")
        if self._lon == None:
            raise Exception("you must specify the lon property")

        #if len(self._circles) > 0:
            #for o in self._circles:
                #self._map.add_layer(o["layers"])

        #if len(self._polygons) > 0:
            #for o in self._polygons:
                #self._map.add_layer(o["layers"])

        self._colors = None

        if self.hasOpt("colormap"):
            self._colors = tools.Colors(colormap=self.getOpt("colormap"))
        elif self.hasOpt("colors"):
            self._colors = tools.Colors(colors=self.getOpt("colors"))
        else:
            self._colors = self._visuals._colors

        if self._colors != None:
            if self.hasOpt("color_range"):
                range = self.getOpt("color_range")
                self._colorRange = tools.ColorRange(self._colors,range[0],range[1])

        components = []

        components.append(self._banner)

        colorbar = None

        if self.hasOpt("color"):
            #data = go.Scatter(x=[None],y=[None],marker=dict(colorscale=self._visuals._colors.colorscale,showscale=True,cmin=-5,cmax=5,colorbar=dict(xpad=0,ypad=0,ticks="")))
            data = go.Scatter(x=[None],y=[None],marker=dict(colorscale=self._visuals._colors.colorscale,showscale=True,colorbar=dict(thickness=30)))
            layout = dict(xaxis=dict(visible=False),yaxis=dict(visible=False),showlegend=False)
            #self._colorbar = go.FigureWidget(data=data,layout=layout)
            #colorbar = widgets.Box([self._colorbar],layout=widgets.Layout(width="150px",margin="0px",padding="0px"))

        if colorbar != None:
            components.append(widgets.HBox([self._map,colorbar]))
        else:
            components.append(self._map)

        self.children = components

    def draw(self,data = None,clear = False):
        if self._map == None:
            return

        popup = self.getValues("popup")

        value = self.getOpt("size")

        tracking = self.getOpt("tracking",False)
        center = None

        sizePixels = None
        sizeField = None

        if value != None: 
            try:
                sizePixels = int(value)
            except:
                sizeField = value

        color = None
        colorField = None

        value = self.getOpt("color")

        if value != None: 
            color = Colors.getColorFromName(value)

            if color == None:
                colorField = value

        minSize = sys.maxsize
        maxSize = 0

        minColor = sys.maxsize
        maxColor = 0

        sizes = self.getOpt("sizes",(1,20))

        sizeRange = []
        colorRange = None

        gradient = None

        data = self._datasource.getList()

        if sizeField != None or colorField != None:
            for value in data:
                if sizeField != None:
                    if sizeField in value:
                        num = float(value[sizeField])
                        if num < minSize:
                            minSize = num
                        if num > maxSize:
                            maxSize = num
                if colorField != None:
                    if colorField in value:
                        num = float(value[colorField])
                        if num < minColor:
                            minColor = num
                        if num > maxColor:
                            maxColor = num
        if sizeField != None:
            if maxSize > minSize:
                sizeRange = np.arange(minSize,maxSize,(maxSize - minSize) / int(sizes[1] - sizes[0]))

        if colorField != None:
            if self._colorRange != None:
                colorRange = self._colorRange
            elif self._colors != None:
                colorRange = tools.ColorRange(self._colors,minColor,maxColor)
            else:
                colorRange = tools.ColorRange(self._visuals._colors,minColor,maxColor)
            if self._colorbar != None:
                self._colorbar.data[0].marker.cmin = minColor
                self._colorbar.data[0].marker.cmax = maxColor
            #baseColor = self.getOpt("base_color",self._visuals._colors.lightest)
            #gradient = tools.Gradient(baseColor,levels=200,min=minColor,max=maxColor)

        opacity = self.getOpt("marker_opacity",1)
        border = self.getOpt("marker_border",True)

        keyValues = self.getValues("keys")

        if len(keyValues) == 0:
            keyValues = ["_key_"]

        keys = []

        for value in data:
            key = ""
            for i,k in enumerate(keyValues):
                if i > 0:
                    key += "."
                if k in value:
                    key += value[k]
                
            keys.append(key)

            if key in self._markers:
                marker = self._markers[key]
            else:
                if self._icon != None:
                    icon = maps.Icon(icon_url=self._icon,icon_size=[40,30])
                    marker = maps.Marker(icon=icon)
                else:
                    marker = maps.CircleMarker()
                    marker.stroke = border
                    marker.fill_opacity = opacity

                self._map.add_layer(marker)
                self._markers[key] = marker
                if len(popup) > 0:
                    marker.popup = widgets.HTML()

            if self._lat in value and self._lon in value:
                lat = value[self._lat]
                lon = value[self._lon]
                marker.location = (lat,lon)

            if tracking and center == None:
                center = marker.location

            if sizePixels != None:
                marker.radius = sizePixels
            elif len(sizeRange) > 0:
                size = 0
                if sizeField in value:
                    size = float(value[sizeField])
                index = np.where(size >= sizeRange)[0]
                marker.radius = len(index) - 1

            if color != None:
                marker.fill_color = color
            elif colorRange != None:
                if colorField in value:
                    num = float(value[colorField])
                    marker.fill_color = colorRange.getColor(num)

            if gradient != None:
                if colorField in value:
                    num = float(value[colorField])
                    marker.fill_color = gradient.darken(num - minColor)

            if marker.popup != None:
                text = ""
                for i,s in enumerate(popup):
                    if i > 0:
                        text += "<br/>"
                    text += s + "=" + value[s]
                marker.popup.value = text

        remove = {}

        for key,marker in self._markers.items():
            if (key in keys) == False:
                remove[key] = marker

        for key,marker in remove.items():
            self._markers.pop(key)
            self._map.remove_layer(marker)

        if center != None:
            self._map.center = center

        self.setTitle()

    def addCircles(self,datasource,**kwargs):
        options = tools.Options(**kwargs)
        o = {}
        o["lat"] = options.getOpt("lat")
        o["lon"] = options.getOpt("lon")
        o["datasource"] = datasource
        o["layers"] = maps.LayerGroup()
        if options.hasOpt("radius"):
            value = options.getOpt("radius")
            try:
                num = int(value)
                o["radius"] = num
            except:
                o["radius_field"] = value
        if options.hasOpt("text"):
            o["text"] = options.getOpt("text")
        self._circles.append(o)
        self.loadCircles(o)
        datasource.addDelegate(self)

    def addPolygons(self,datasource,**kwargs):
        options = tools.Options(**kwargs)
        o = {}
        o["coords"] = options.getOpt("coords")
        o["datasource"] = datasource
        o["layers"] = maps.LayerGroup()
        if options.hasOpt("text"):
            o["text"] = options.getOpt("text")
        o["order"] = options.getOpt("order","lat_lon")
        self._polygons.append(o)
        self.loadPolygons(o)
        datasource.addDelegate(self)

    def dataChanged(self,datasource,data,clear):
        o = None

        for circle in self._circles:
            if circle["datasource"] == datasource:
                o = circle
                break

        if o != None:
            self.loadCircles(o)
        else:
            for polygon in self._polygons:
                if polygon["datasource"] == datasource:
                    o = polygon
                    break
            if o != None:
                self.loadPolygons(o)
    
    def loadCircles(self,o):
        o["layers"].clear_layers()

        data = o["datasource"].getData()

        for key,value in data.items():
            lat = value[o["lat"]]
            lon = value[o["lon"]]

            radius = 50

            if "radius" in o:
                radius = o["radius"]
            elif "radius_field" in o:
                num = float(value[o["radius_field"]])
                radius = int(num)

            circle = maps.Circle(location=(lat,lon),radius=radius)

            borderWidth = self.getOpt("circle_border_width",2)
            circle.weight = borderWidth

            if borderWidth > 0:
                circle.stroke = True
                circle.color = self.getOpt("circle_border_color","black")
            else:
                circle.stroke = False

            circle.fill_color = self.getOpt("circle_fill_color","white")
            circle.fill_opacity = self.getOpt("circle_fill_opacity",.2)

            if "text" in o:
                text = value[o["text"]]
                circle.popup = widgets.HTML(value=text)
            o["layers"].add_layer(circle)

        if self._map != None:
            if (o["layers"] in self._map.layers) == False:
                self._map.add_layer(o["layers"])
 
    def loadPolygons(self,o):
        o["layers"].clear_layers()

        data = o["datasource"].getData()

        lonlat = (o["order"] == "lon_lat")

        for key,value in data.items():
            coords = value[o["coords"]]
            a = coords.split(" ")
            points = []
            i = 0
            while True:
                if lonlat:
                    points.append((a[i + 1],a[i]))
                else:
                    points.append((a[i],a[i + 1]))
                i += 2
                if i >= len(a):
                    break
            polygon = maps.Polygon(locations=points)

            borderWidth = self.getOpt("poly_border_width",1)
            polygon.weight = borderWidth

            if borderWidth > 0:
                polygon.stroke = True
                polygon.color = self.getOpt("poly_border_color","black")
            else:
                polygon.stroke = False

            polygon.fill_color = self.getOpt("poly_fill_color","white")
            polygon.fill_opacity = self.getOpt("poly_fill_opacity",.2)

            if "text" in o:
                text = value[o["text"]]
                polygon.popup = widgets.HTML(value=text)

            o["layers"].add_layer(polygon)

        if self._map != None:
            if (o["layers"] in self._map.layers) == False:
                self._map.add_layer(o["layers"])

    @property
    def map(self):
        return(self._map)

    @property
    def icon(self):
        return(self._icon)

    @icon.setter
    def icon(self,value):
        self._icon = value

class Gauge(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)
        self._intervalColors = None
        self._segments = 0
        self._gauges = {}
        self.range = self.getOpt("range",(0,100))

    def createContent(self):
        self._segments = self.getOpt("segments",3)

        self._shape = self.getOpt("shape",50)

        if self._shape >= 90:
            raise Exception("the shape value must be less than 90")

        interval = self._shape / self._segments
        mod = self._shape % self._segments

        self._intervalColors = [self.getOpt("label_bg","rgba(0,0,0,0)")]

        if self.hasOpt("colors"):
            self._intervalColors.extend(self.getOpt("colors"))
        else:
            if self.hasOpt("color"):
                color = self.getOpt("color")
            else:
                color = self._visuals._colors.lightest

            for i in range(0,self._segments):
                self._intervalColors.append(tools.darken(color,(i * 20)))

        self._intervalValues = [100 - self._shape]
        self._intervalLabels = [" "]
        self._ticValues = [100 - self._shape - 10]
        self._ticLabels = [" "]

        self._segmentLabels = [self.getOpt("label"," ")]

        text = self.getOpt("text",[])

        intervalSize = (self._range[1] - self._range[0]) / self._segments

        self._intervals = np.arange(self._range[0],self._range[1],intervalSize)

        for i in range(0,self._segments):
            value = int(interval)
            if mod > 0:
                value += 1
                mod -= 1
            self._intervalValues.append(value)
            self._intervalLabels.append("i" + str(i))

            self._ticValues.append(value)

            self._ticLabels.append(self.getLabel(self._range[0] + (intervalSize * i)))

            if len(text) > i:
                self._segmentLabels.append(text[i])
            else:
                self._segmentLabels.append(" ")

            if i == self._segments - 1:
                if mod > 0:
                    value += 1
                self._ticValues.append(value)
                self._ticLabels.append(self.getLabel(self._range[0] + (intervalSize * (i + 1))))

        self.children = [self._banner]

    def draw(self,data = None,clear = False):

        if data == None or len(data) == 0:
            data = self._datasource.getList()

        gauges = None

        field = self.getOpt("value")

        if field == None:
            return

        items = None

        currentLength = len(self._gauges)

        layout = False

        if clear:
            self._gauges = {}
            layout = True

        for o in data:
            if ("_key_" in o) == False:
                continue
            key = o["_key_"]

            if "_opcode" in o and o["_opcode"] == "delete":
                if key in self._gauges:
                    self._gauges.pop(key)
                    layout = True
            else:
                if (key in self._gauges) == False:
                    gauge = GaugeEntry(self)
                    gauge.setOpt("title",key)
                    self._gauges[key] = gauge
                    layout = True
                else:
                    gauge = self._gauges[key]

                value = float(o[field])
                gauge.value = value

        if layout:
            columns = self.getOpt("columns",3)
            a = [self._banner]
            g = []
            column = 0
            for key,value in self._gauges.items():
                if column < columns:
                    g.append(value)
                    column += 1
                else:
                    a.append(widgets.HBox(g,layout=widgets.Layout(justify_content="center")))
                    g = [value]
                    column = 1

            if len(g) > 0:
                a.append(widgets.HBox(g,layout=widgets.Layout(justify_content="center")))

            if self.getOpt("show_controls",False):
                if self._controls == None:
                    self._controls = ControlPanel(self._datasource) 
                a.append(self._controls)

            self.children = a

        if len(self._gauges) > 1:
            if currentLength <= 1:
                for gauge in self._gauges.values():
                    gauge.setTitle()
        elif len(self._gauges) == 1:
            if currentLength > 1:
                for gauge in self._gauges.values():
                    gauge.setTitle()

        self.setTitle()

    def setTitle(self,title = None):
        title = self.getOpt("title","")

        if len(self._gauges) == 1:
            for gauge in self._gauges.values():
                if len(title) == 0:
                    title = gauge.getOpt("title","")
                title = str.format("{} ({})".format(title,int(gauge._value)))
                break
            
        Chart.setTitle(self,title)

    def getLabel(self,num):
        if num == 0:
            label = "0"
        elif num >= 1000000:
            label = str(num / 1000000) + "M"
        elif num >= 100000:
            label = str(int(num / 10000)) + "K"
        elif self._range[1] - self._range[0] < 10:
            label = str(num)
        else:
            label = str(int(num))

        return(label)

    @property
    def range(self):
        return(self._range)

    @range.setter
    def range(self,value):
        if value[1] > value[0]:
            self._range = value
            self.create()

class GaugeEntry(Options,widgets.VBox):
    def __init__(self,gauge,**kwargs):
        Options.__init__(self,**kwargs)
        widgets.VBox.__init__(self,layout=widgets.Layout(border=gauge.getOpt("border","1px solid #d8d8d8"),padding=gauge.getOpt("padding","10px"),margin=gauge.getOpt("spacing","10px")))
        self._gauge = gauge
        self._figure = None
        self._layout = None
        self._value = 0
        self._title = widgets.HTML()

    def create(self):
        size = self._gauge.getOpt("size",300)

        self._layout = go.Layout(width=size,height=size)

        self._data = []

        self._tics = go.Pie(values=self._gauge._ticValues,labels=self._gauge._ticLabels,
            marker=dict(colors=["rgba(0,0,0,0)"] * (self._gauge._segments + 10),line_width = 0),
            direction="clockwise",
            rotation=int((((self._gauge._shape + 10) / 100) * 360) / 2),
            hole=self._gauge.getOpt("center_size",.40),
            sort=False,
            showlegend=False,
            hoverinfo="none",
            textposition="outside",
            textinfo="label")

        self._data.append(self._tics)

        self._intervals = go.Pie(values=self._gauge._intervalValues,
            labels=self._gauge._intervalLabels,
            text=self._gauge._segmentLabels,
            marker=dict(line_width=self._gauge.getOpt("line_width",4)),
            marker_colors=self._gauge._intervalColors,
            hole=self._gauge.getOpt("center_size",.40),
            sort=False,
            direction="clockwise",
            rotation=int(((self._gauge._shape / 100) * 360) / 2),
            showlegend=False,
            hoverinfo="none",
            textposition="inside",
            textinfo="text")

        self._data.append(self._intervals)

        margin = self._gauge.getOpt("margin",30)
        self._layout["margin"] = dict(l=margin,r=margin,b=margin,t=margin)

        #self._layout["paper_bgcolor"] = "#e8e8e8"
        #self._layout["plot_bgcolor"] = "blue"
        #self._layout["paper_bgcolor"] = "rgba(0,0,0,0)"
        self._layout["plot_bgcolor"] = "rgba(0,0,0,0)"

        self._figure = go.FigureWidget(data=self._data,layout=self._layout)
 
        height = size
        height += 30
        #self.children = [self._title,self._figure],layout=widgets.Layout(border="1px solid #d8d8d8",width=str(size) + "px",height=str(height) + "px")
        self.children = [self._title,self._figure]
 
    def draw(self,data = None,clear = False):

        if self._figure == None:
            self.create()

        num = (self._value - self._gauge._range[0])  / (self._gauge._range[1] - self._gauge._range[0])

        angleRange = (self._gauge._shape / 100) * 360

        start = angleRange - ((angleRange - 180) / 2)

        angle = start - (angleRange * num)

        hole = self._intervals["hole"]

        radians = angle * (math.pi / 180)

        radius = hole / 2
        radius -= .02

        center = dict(x=.5,y=.5)

        if self._intervals.domain.x != None:
            center["x"] = float((self._intervals.domain.x[1] - self._intervals.domain.x[0]) / 2)

        if self._intervals.domain.y != None:
            center["y"] = float((self._intervals.domain.y[1] - self._intervals.domain.y[0]) / 2)

        x = center["x"] + (radius * math.cos(radians))
        y = center["y"] + (radius * math.sin(radians))

        r = .05

        radians = (angle - 75) * (math.pi / 180)
        x0 = center["x"] + (r * math.cos(radians))
        y0 = center["y"] + (r * math.sin(radians))

        radians = (angle + 75) * (math.pi / 180)
        x1 = center["x"] + (r * math.cos(radians))
        y1 = center["y"] + (r * math.sin(radians))

        path = ""
        path += "M " + str(x0) + " " + str(y0)
        path += "L " + str(x) + " " + str(y)
        path += "L " + str(x1) + " " + str(y1)

        interval = np.where(self._value >= self._gauge._intervals)[0]

        fillcolor = self._gauge._intervalColors[len(interval)]

        hole = hole / 2

        centerBg = self._gauge.getOpt("center_bg","rgba(0,0,0,0)")

        linewidth = self._gauge.getOpt("line_width",4)

        centerCircle = go.layout.Shape(type="circle",x0=center["x"] - hole,y0=center["y"] - hole,x1=center["x"] + hole,y1=center["y"] + hole,fillcolor=centerBg,line_width=linewidth)

        lw = linewidth
        if lw == 0:
            lw = 2

        pointerCircle = go.layout.Shape(type="circle",x0=center["x"] - r,y0=center["y"] - r,x1=center["x"] + r,y1=center["y"] + r,fillcolor=fillcolor,line_width=lw)
        pointer = go.layout.Shape(type="path",path=path,fillcolor=fillcolor,line_width=lw)

        shapes = [centerCircle,pointerCircle,pointer]

        #label = go.layout.Annotation(text=str(int(self._value)),x=center["x"],y=center["y"] - (r * 2),showarrow=False,font=dict(size=14))
        #label = go.layout.Annotation(text=str(int(self._value)),x=center["x"],y=center["y"] - .4,showarrow=False,font=dict(size=18))

        #self._figure.update_layout(shapes=shapes,annotations=[label])
        self._figure.update_layout(shapes=shapes)

        self.setTitle()

    def setTitle(self):
        if len(self._gauge._gauges) > 1:
            title = self.getOpt("title")

            if title != None:
                self._title.value = self._gauge._visuals.formatTitle("{} ({})".format(title,int(self._value)))

            self.layout.border = self._gauge.getOpt("border","1px solid #d8d8d8")
        else:
            self._title.value = ""
            self.layout.border = "0"

    @property
    def value(self):
        return(self._value)

    @value.setter
    def value(self,value):
        self._value = value
        self.draw()

class Compass(Chart):
    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)
        self._colors = None
        self._entries = {}

    def createContent(self):
        segments = 8

        self._values = [13,12,13,12,13,12,13,12]
        self._labels = ["N","NE","E","SE","S","SW","W","NW"]

        self.children = [self._banner]

    def draw(self,data = None,clear = False):

        if data == None or len(data) == 0:
            data = self._datasource.getList()

        field = self.getOpt("heading")

        if field == None:
            return

        layout = False

        if clear:
            self._entries = {}
            layout = True

        for o in data:
            if ("_key_" in o) == False:
                continue

            key = o["_key_"]

            if "_opcode" in o and o["_opcode"] == "delete":
                if key in self._entries:
                    self._entries.pop(key)
                    layout = True
            else:
                if (key in self._entries) == False:
                    entry = CompassEntry(self)
                    entry.setOpt("title",key)
                    self._entries[key] = entry
                    layout = True
                else:
                    entry = self._entries[key]

                value = float(o[field])
                entry.heading = value

        if layout:
            columns = self.getOpt("columns",3)
            a = [self._banner]
            g = []
            column = 0
            for key,value in self._entries.items():
                if column < columns:
                    g.append(value)
                    column += 1
                else:
                    a.append(widgets.HBox(g,layout=widgets.Layout(justify_content="center")))
                    g = [value]
                    column = 1

            if len(g) > 0:
                a.append(widgets.HBox(g,layout=widgets.Layout(justify_content="center")))

            self.children = a

        self.setTitle()

    def setTitle(self,title = None):
        title = self.getOpt("title","")

        if len(self._entries) == 1:
            for entry in self._entries.values():
                if len(title) == 0:
                    title = entry.getOpt("title","")
                title = str.format("{} ({})".format(title,int(entry._heading)))
                break
            
        Chart.setTitle(self,title)

class CompassEntry(Options,widgets.VBox):
    def __init__(self,compass,**kwargs):
        Options.__init__(self,**kwargs)
        widgets.VBox.__init__(self,layout=widgets.Layout(border=compass.getOpt("border","1px solid #d8d8d8"),padding=compass.getOpt("padding","10px"),margin=compass.getOpt("margin","10px")))
        self._compass = compass
        self._figure = None
        self._layout = None
        self._heading = 0
        self._title = widgets.HTML()
        self._innerCircle = None

    def create(self):
        size = self._compass.getOpt("size",300)

        self._layout = go.Layout(width=size,height=size)

        self._data = []

        color = Colors.getColorFromName(self._compass.getOpt("outer_color","#89cff0"))

        linewidth = self._compass.getOpt("line_width",1)

        self._intervals = go.Pie(values=self._compass._values,
            #domain={"x":[0,.5]},
            labels=self._compass._labels,
            marker=dict(colors=[color] * 20,line_width=self._compass.getOpt("line_width",linewidth)),
            hole=self._compass.getOpt("compass_size",.90),
            sort=False,
            direction="clockwise",
            showlegend=False,
            hoverinfo="none",
            textposition="outside",
            textinfo="none")

        self._data.append(self._intervals)

        margin = self._compass.getOpt("margin",30)
        self._layout["margin"] = dict(l=margin,r=margin,b=margin,t=margin)

        labels = []

        headings = ["N", "NE", "E", "SE", "S", "SW", "W", "NW"]

        hole = self._intervals["hole"]
        radius = hole / 2
        center = dict(x=.5,y=.5)

        color = self._compass.getOpt("bg_color","#f8f8f8")

        textangle = 0
        angle = 0

        for i,text in enumerate(headings):
            if angle == 135 or angle == 315:
                h = (450 - angle - 3) % 360
            else:
                h = (450 - angle) % 360

            radians = h * (math.pi / 180)
            if i % 2 == 1:
                fontsize = 12
                length = radius - .08
            else:
                fontsize = 18
                length = radius - .1
            x = center["x"] + (length * math.cos(radians))
            y = center["y"] + (length * math.sin(radians))
            label = go.layout.Annotation(text=headings[i],x=x,y=y,showarrow=False,font=dict(size=fontsize),textangle=textangle)
            labels.append(label)
            angle += (360 / len(headings))
            textangle += 45

        textangle = 0
        angle = 0

        for i in range(0,8):
            if angle == 135 or angle == 315:
                h = (450 - angle - 2) % 360
            else:
                h = (450 - angle) % 360
            radians = h * (math.pi / 180)
            if i % 2 == 1:
                length = radius + .03
            else:
                length = radius
            x = center["x"] + (length * math.cos(radians))
            y = center["y"] + (length * math.sin(radians))
            label = go.layout.Annotation(text=str(angle),x=x,y=y,showarrow=False,font=dict(size=10),textangle=textangle)
            labels.append(label)
            angle += 45
            textangle += 45

        h = (hole / 2) - .1
        self._innerCircle = go.layout.Shape(type="circle",x0=center["x"] - h,y0=center["y"] - h,x1=center["x"] + h,y1=center["y"] + h,fillcolor=color,line_width=linewidth)

        self._layout["annotations"] = labels
        self._layout["shapes"] = [self._innerCircle]

        self._figure = go.FigureWidget(data=self._data,layout=self._layout)
 
        height = size
        height += 30
        #self.children = widgets.VBox([self._title,self._figure],layout=widgets.Layout(border="1px solid #d8d8d8",width=str(size) + "px",height=str(height) + "px"))
        self.children = [self._title,self._figure]
 
    def draw(self,data = None,clear = False):
        if self._figure == None:
            self.create()

        heading = (450 - self._heading) % 360
        hole = self._intervals["hole"]
        radius = hole / 2

        center = dict(x=.5,y=.5)

        if self._intervals.domain.x != None:
            center["x"] = float((self._intervals.domain.x[1] - self._intervals.domain.x[0]) / 2)

        if self._intervals.domain.y != None:
            center["y"] = float((self._intervals.domain.y[1] - self._intervals.domain.y[0]) / 2)
        linewidth = self._compass.getOpt("line_width",1)

        r = .05

        # Heading Pointer

        radians = heading * (math.pi / 180)

        length = radius - .1

        x = center["x"] + (length * math.cos(radians))
        y = center["y"] + (length * math.sin(radians))

        radians = (heading - 75) * (math.pi / 180)
        x0 = center["x"] + (r * math.cos(radians))
        y0 = center["y"] + (r * math.sin(radians))

        radians = (heading + 75) * (math.pi / 180)
        x1 = center["x"] + (r * math.cos(radians))
        y1 = center["y"] + (r * math.sin(radians))

        path = ""
        path += "M " + str(x0) + " " + str(y0)
        path += "L " + str(x) + " " + str(y)
        path += "L " + str(x1) + " " + str(y1)

        color = self._compass.getOpt("heading_color","#89cff0")
        headingPointer = go.layout.Shape(type="path",path=path,fillcolor=color,line_width=linewidth)

        # End Heading Pointer

        # Opposite Pointer

        radians = ((heading + 180) % 360) * (math.pi / 180)

        length = radius - .1

        x = center["x"] + (length * math.cos(radians))
        y = center["y"] + (length * math.sin(radians))

        radians = (heading - 75) * (math.pi / 180)
        x0 = center["x"] + (r * math.cos(radians))
        y0 = center["y"] + (r * math.sin(radians))

        radians = (heading + 75) * (math.pi / 180)
        x1 = center["x"] + (r * math.cos(radians))
        y1 = center["y"] + (r * math.sin(radians))

        path = ""
        path += "M " + str(x0) + " " + str(y0)
        path += "L " + str(x) + " " + str(y)
        path += "L " + str(x1) + " " + str(y1)

        color = self._compass.getOpt("reciprocal_color","white")
        reciprocalPointer = go.layout.Shape(type="path",path=path,fillcolor=color,line_width=linewidth)

        # End Opposite Pointer

        rr = r + .02
        color = self._compass.getOpt("center_color","white")
        needleCenter = go.layout.Shape(type="circle",x0=center["x"] - rr,y0=center["y"] - rr,x1=center["x"] + rr,y1=center["y"] + rr,fillcolor=color,line_width=linewidth)

        shapes = [headingPointer,reciprocalPointer]
        shapes = [self._innerCircle,headingPointer,reciprocalPointer,needleCenter]
        labels = []
        label = go.layout.Annotation(text=str(int(self._heading)),x=center["x"],y=center["y"],showarrow=False,font=dict(size=12))
        #labels.append(label)

        self._figure.update_layout(shapes=shapes,annotations=labels)

        self.setTitle()

    def setTitle(self):
        title = self.getOpt("title")

        if len(self._compass._entries) > 1:
            title = self.getOpt("title")

            if title != None:
                self._title.value = self._compass._visuals.formatTitle("{} ({})".format(title,int(self._heading)))

            self.layout.border = self._compass.getOpt("border","1px solid #d8d8d8")
        else:
            self.layout.border = "0"

    @property
    def heading(self):
        return(self._heading)

    @heading.setter
    def heading(self,value):
        self._heading = value
        self.draw()

class ControlPanel(widgets.HBox):
    def __init__(self,datasource):

        widgets.HBox.__init__(self)

        self._datasource = datasource

        self._info = None

        if isinstance(self._datasource,connections.EventCollection):
            self._info = self._datasource.getInfo()

        components = []

        self._filter = widgets.Text(description="Filter",layout=widgets.Layout(width="50%"))
        b = widgets.Button(description="Set",layout=widgets.Layout(width="50px"))
        b.on_click(self.filter)

        self._buttons = None

        self._playPauseButton = widgets.Button(icon="fa-pause",layout=widgets.Layout(width="40px",margin="0px 0px 0px 20px"))
        self._playPauseButton.on_click(self.playPause)

        if self._datasource._paused:
            self._playPauseButton.icon = "fa-play"

        if isinstance(self._datasource,connections.EventCollection):
            iconLayout = widgets.Layout(width="40px")

            self._nextButton = widgets.Button(icon="fa-step-forward",layout=iconLayout)
            self._prevButton = widgets.Button(icon="fa-step-backward",layout=iconLayout)
            self._firstButton = widgets.Button(icon="fa-backward",layout=iconLayout)
            self._lastButton = widgets.Button(icon="fa-forward",layout=iconLayout)

            self._nextButton.on_click(self.next)
            self._prevButton.on_click(self.prev)
            self._firstButton.on_click(self.first)
            self._lastButton.on_click(self.last)

            self._buttons = widgets.HBox([self._nextButton,self._prevButton,self._firstButton,self._lastButton,self._playPauseButton],layout=widgets.Layout(width="300px"))

        else:
            self._buttons = widgets.HBox([self._playPauseButton],layout=widgets.Layout(width="100px"))

            #if self._datasource._pages == 1:
                #self._buttons.layout.display = "none"
            #else:
                #self._buttons.layout.display = "block"

        components.append(self._buttons)

        w = "80%"
        justify = "flex-end"
        if self._info == None or self._datasource._pages == 1:
            w = "100%"
            justify = "center"
        self._filterBox = widgets.HBox([self._filter,b],layout=widgets.Layout(width=w,justify_content=justify))
        components.append(self._filterBox)

        #self._panel = widgets.HBox(components,layout=widgets.Layout(width="100%",border="1px solid #d8d8d8"))

        self._filter.value = self._datasource.getFilter()

        self.children = components

    def next(self,b):
        self._datasource.next()

    def prev(self,b):
        self._datasource.prev()

    def first(self,b):
        self._datasource.first()

    def last(self,b):
        self._datasource.last()

    def playPause(self,b):
        code = self._datasource.togglePlay()
        if code:
            self._playPauseButton.icon = "fa-pause"
        else:
            self._playPauseButton.icon = "fa-play"

    def filter(self,b):
        self._datasource.setFilter(self._filter.value)
        self._datasource.load()

    def processInfo(self):
        if isinstance(self._datasource,connections.EventCollection):
            #if self._datasource._pages == 1:
                #self._buttons.layout.display = "none"
            #else:
                #self._buttons.layout.display = "block"

            self._firstButton.disabled = (self._datasource._pages == 1)
            self._lastButton.disabled = (self._datasource._pages == 1)
            self._nextButton.disabled = (self._datasource._page == (self._datasource._pages - 1))
            self._prevButton.disabled = (self._datasource._page == 0)

            if self._datasource._paused:
                self._playPauseButton.icon = "fa-play"
            else:
                self._playPauseButton.icon = "fa-pause"

class Controls(Chart):

    def __init__(self,visuals,datasource,**kwargs):
        Chart.__init__(self,visuals,datasource,**kwargs)
        self._panel = ControlPanel(self._datasource)
        self.children = [self._banner,self._panel]

    def draw(self,data,clear):
        pass

    def info(self):
        self._panel.processInfo()
        self.setTitle()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Utility for handling authentication info

'''

from base64 import b64encode

class Authorization(object):
    _instance = None

    @staticmethod
    def getInstance(session):
        if Authorization._instance == None:
            Authorization._instance = Authorization(session)

        return(Authorization._instance)

    def __init__(self,session):
        self._session = session
        self._authorization = None
        self._hostname = None

    def setBasic(self,username,password):
        basic = username + ":" + password
        encoded = b64encode(basic.encode())
        self._authorization = "Basic " + encoded.decode()

    def setBearer(self,token):
        self._authorization = "Bearer " + token

    def setKerberos(self,hostname):
        self._hostname = hostname

    @property
    def isEnabled(self):
        return(self._authorization != None or self._hostname != None)

    @property
    def authorization(self):
        authorization = None

        if self._hostname != None:
            authorization = self._session.auth.generate_request_header(None,self._hostname,True)
        else:
            authorization = self._authorization

        return(authorization)




#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP MAS Modules '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
from six.moves import urllib
from .base import ESPObject
from .utils import xml
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name


class ModuleMember(object):

    def __init__(self, member, sha_key, type, description=None,
                 code_file=None, code=None):
        self.member = member
        self.sha_key = sha_key
        self.type = type
        self.description = description
        self.code_file = code_file
        self.code = code

    @classmethod
    def from_element(cls, data, session=None):
        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        out = cls(data.attrib['member'], data.attrib['SHAkey'],
                  data.attrib['type'])

        for item in data.findall('./description'):
            out.description = item.text

        for item in data.findall('./code'):
            out.code = item.text
        for item in data.findall('./code-file'):
            out.code_file = item.text

        return out

    from_xml = from_element

    def to_element(self):
        out = xml.new_elem('module-member',
                           attrib=dict(member=self.member,
                                       SHAkey=self.sha_key,
                                       type=self.type))

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        if self.code:
            xml.add_elem(out, 'code', text_content=self.code)
        if self.code_file:
            xml.add_elem(out, 'code-file', text_content=self.code_file)

        return out

    def to_xml(self, pretty=False):
        return xml.to_xml(self.to_element(), pretty=pretty)


class MASModule(ESPObject):
    '''
    MAS Module

    Attributes
    ----------
    module : string
        Name of the MAS module
    project : string
        Name of the project the module belongs to

    Parameters
    ----------
    language : string
        The name of the programming language
    func_names : string or list-of-strings, optional
        The function names exported by the module
    module : string, optional
        Name of the MAS module

    '''

    def __init__(self, language, module, func_names, mas_store=None,
                 mas_store_version=None, description=None, code_file=None,
                 code=None):
        ESPObject.__init__(self)
        self.module = module or gen_name('mas_')
        self.language = language
        if isinstance(func_names, six.string_types):
            func_names = re.split(r'\s*,\s*', func_names.strip())
        if isinstance(func_names, six.string_types):
            self.func_names = re.split(r'\s*,\s', func_names.strip())
        else:
            self.func_names = list(func_names)
        self.mas_store = mas_store
        self.mas_store_version = mas_store_version
        self.description = description
        self.code_file = code_file
        self.code = code
        self.module_members = []
        self.project = None

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create MAS module from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML MAS module definition
        session : requests.Session, optional
            Session that the MAS module is associated with

        Returns
        -------
        :class:`MASModule`

        '''
        if isinstance(data, six.string_types):
            data = xml.from_xml(get_project_data(data))

        out = cls(data.attrib['language'], data.attrib['module'],
                  func_names=data.attrib['func-names'],
                  mas_store=data.attrib.get('mas-store'),
                  mas_store_version=data.attrib.get('mas-store-version'))

        out.session = session

        for item in data.findall('./code-file'):
            out.code_file = item.text
        for item in data.findall('./code'):
            out.code = item.text
        for item in data.findall('./description'):
            out.description = item.text

        for item in data.findall('./module-members/module-member'):
            out.module_members.append(ModuleMember.from_element(item, session=session))

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Convert MAS module to ElementTree.Element definition

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('mas-module',
                           attrib=dict(module=self.module,
                                       language=self.language,
                                       func_names=','.join(self.func_names),
                                       mas_store=self.mas_store,
                                       mas_store_version=self.mas_store_version))

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        if self.code:
            xml.add_elem(out, 'code', text_content=self.code)
        elif self.code_file:
            xml.add_elem(out, 'code-file', text_content=self.code_file)

        if self.module_members:
            members = xml.add_elem(out, 'module-members')
            for item in self.module_members:
                xml.add_elem(members, item.to_element())

        return out

    def to_xml(self, pretty=False):
        '''
        Convert MAS module to XML definition

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def save(self, path=None):
        '''
        Save the module

        Parameters
        ----------
        path : string, optional
            The location to save the module

        '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'masModules/%s/%s/state' % (self.project, self.module)),
                  params=get_params(value='persisted', path=path))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Plotting Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals

import base64
import collections
import copy
import csv
import datetime
import io
import json
import numpy as np
import os
import pandas as pd
import re
import requests
import six
import threading
import time
import uuid
import warnings
import weakref
import xml.etree.ElementTree as ET
from six.moves import urllib
from .base import ESPObject
from .config import CONCAT_OPTIONS
from .schema import Schema
from .utils import xml
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name
from .utils.events import get_events

#
# The streaming_* plot methods create a chart automatically so they
# must accept parameters for both the chart and for the plotting methods.
# This is the list of chart parameters that get filtered out.
#
CHART_PARAMS = set(re.split(r'\s+', '''
    title x_axis_label x_axis_location x_axis_type y_axis_label
    y_axis_location y_axis_type x_range y_range plot_width plot_height
    responsive palette
'''.strip()))


def listify(data):
    ''' Make sure data is a list '''
    if data is None:
        return []
    if isinstance(data, (set, list, tuple)):
        return list(data)
    return [data]


def camelize(name):
    ''' Convert underscore delimited name to camel-case '''
    return re.sub(r'[_-](\w)', lambda m: m.group(1).upper(), name)


def split_chart_params(**kwargs):
    '''
    Split the keyword arguments into chart and plot method params

    Parameters
    ----------
    **kwargs : keyword arguments
        The parameters to split into categories

    Returns
    -------
    (chart-params-dict, plot-method-params-dict)

    '''
    chart_params = {}
    mthd_params = {}
    for key, value in kwargs.items():
        if key in CHART_PARAMS:
            chart_params[key] = value
        else:
            mthd_params[key] = value
    return chart_params, mthd_params


def highlight_image(pil_image, detections, line_width=3, font=None):
    from PIL import ImageDraw

    row = detections.iloc[-1]
    n_objects = row['_nObjects_']
    if n_objects == 0:
        return pil_image

    draw = ImageDraw.Draw(pil_image)

    for i in range(0, int(n_objects)):
        obj = row['_Object%s_' % i]
        prob = row['_P_Object%s_' % i]
        x = int(row['_Object%s_x' % i])
        y = int(row['_Object%s_y' % i])
        width = int(row['_Object%s_width' % i])
        height = int(row['_Object%s_height' % i])
        value = int(prob * 255)

        points = (
            (x, y),
            (x+width, y),
            (x+width, y+height),
            (x, y+height),
            (x, y)
        )

        color = (value, 0, 0, 128)

        draw.line(points, fill=color, width=line_width)
        for point in points:
            draw.ellipse((point[0] - int(line_width / 2),
                          point[1] - int(line_width / 2),
                          point[0] + int(line_width / 2),
                          point[1] + int(line_width / 2)),
                          outline=color, fill=color)

        #draw.rectangle([(x, y), (x+width, y+height)],
        #               outline=(value, 0, 0, 192),
        #               fill=None)
        #font = ImageFont.truetype('/usr/share/fonts/dejavu/DejaVuSans.ttf', 16)
        draw.text((x, y), obj, fill=(255, 255, 255, 255), font=font)
        #cv2.rectangle(img_np, (x, y), (x+width,y+height), (value,0, 0), 2)
        #cv2.putText(img_np, obj, (x,y), cv2.FONT_HERSHEY_SIMPLEX, 1, (255,255,255), 2)

    return pil_image


# Python 2 / 3 compatibility
try:
    long
except:
    long = int


class ESPDataEncoder(json.JSONEncoder):
    '''
    Custom JSON encoder for ESP data

    '''

    def default(self, obj):
        '''
        Convert objects unrecognized by the default encoder

        Parameters
        ----------
        obj : any
           Arbitrary object to convert

        Returns
        -------
        any
           Python object that JSON encoder will recognize

        '''
        if isinstance(obj, (float, np.float64)):
            return float(obj)
        if isinstance(obj, (long, np.int64)):
            return long(obj)
        if isinstance(obj, (int, np.int32)):
            return int(obj)
        return json.JSONEncoder.default(self, obj)


def dump_json(*args, **kwargs):
    '''
    Encode JSON using ESP data types

    '''
    kwargs['cls'] = ESPDataEncoder
    return json.dumps(*args, **kwargs)


class StreamingChart(object):
    '''
    Streaming Chart

    Parameters
    ----------
    data : Window or DataFrame or callable
        The object to get the data from
    title : string, optional
        The title for the chart
    x_axis_label : string, optional
        The label for the x-axis
    x_axis_location : string, optional
        The location of the x-axis: 'top', 'bottom'
    x_axis_type : string, optional
        The type of x-axis: 'linear', 'category', 'logarithmic'
    y_axis_label : string, optional
        The label for the y-axis
    y_axis_location : string, optional
        The location of the y-axis: 'left', 'right'
    y_axis_type : string, optional
        The type of y-axis: 'linear', 'category', 'logarithmic'
    x_range : tuple, optional
        The range and step size of the x-axis in the form (min, max, step)
    y_range : tuple, optional
        The range and step size of the y-axis in the form (min, max, step)
    plot_width : int, optional
        The width of the plot
    plot_height : int, optional
        The height of the plot
    responsive : bool, optional
        Should the dimensions of the plot adjust to fit?
    palette : callable, optional
        Callable object to return a specified number of colors
    steps : int, optional
        The maximum number of calls to update the data in the chart
    interval : int, optional
        The time between data calls in milliseconds
    max_data : int, optional
        The maximum number of data points to store in each data set
    var_generator : callable, optional
        Callable object used to create new transient data columns for
        use in chart parameters

    Returns
    -------
    :class:`StreamingChart`

    '''

    def __init__(self, data, title=None,
                 x_axis_label=None, x_axis_location=None, x_axis_type=None,
                 y_axis_label=None, y_axis_location=None, y_axis_type=None,
                 x_range=None, y_range=None, plot_width=900, plot_height=400,
                 responsive=True, palette=None,
                 steps=1e5, interval=1000, max_data=100, var_generator=None):
        self.max_data = max(int(max_data), 1)
        self.steps = max(int(steps), 0)
        self.interval = max(int(interval), 10)
        self.plot_width = max(plot_width, 10)
        self.plot_height = max(plot_height, 10)

        if palette is None:
            def default_palette(num):
                return ['#1f78b4', '#ff7f00', '#33a02c', '#6a3d9a', '#e31a1c',
                        '#a6cee3', '#cab2d6', '#b2df8a', '#cab2d6', '#fb9a99',
                        '#ffff99', '#b15928'][:num]
            self.palette = default_palette
        else:
            self.palette = palette

        self._data_callback = self._make_callback(data, max_data=self.max_data)
        self._var_generator = var_generator

        self.type = 'line'

        self.data = dict(
            labels=[],
            datasets=[],
        )

        self.options = dict(
            responsive=responsive,

            title=dict(
                display=title and True or False,
                text=title or ''
            ),

            legend=dict(
                display=True,
                position='right',
                labels=dict(
                    usePointStyle=True
                )
            ),

        )

        if [x for x in [x_range, y_range,
                        x_axis_label, y_axis_label,
                        x_axis_type, y_axis_type] if x is not None]:
            scales = dict(xAxes=[{}], yAxes=[{}])

            self.options['scales'] = scales

            if x_axis_type is not None:
                scales['xAxes'][0]['type'] = x_axis_type
            if y_axis_type is not None:
                scales['yAxes'][0]['type'] = y_axis_type

            if x_axis_label is not None:
                scales['xAxes'][0].setdefault('scaleLabel', {})['labelString'] = \
                    x_axis_label
                scales['xAxes'][0].setdefault('scaleLabel', {})['display'] = True
            if y_axis_label is not None:
                scales['yAxes'][0].setdefault('scaleLabel', {})['labelString'] = \
                    y_axis_label
                scales['yAxes'][0].setdefault('scaleLabel', {})['display'] = True

            self._set_range(x_range, 'x')
            self._set_range(y_range, 'y')

    def _set_range(self, limits, axis, index=0):
        '''
        Set range limits

        Parameters
        ----------
        limits : list or tuple
            The min, max, and step-size of the axis
        axis : string
            Which axis to apply range to: 'x' or 'y'
        index : int, optional
            Which axis index to apply range to (if using multiple axes)

        '''
        limits = listify(limits)
        if limits is not None:
            limits += [None, None, None]
            ticks = self.options['scales']['%sAxes' % axis][index].setdefault('ticks', {})
            if limits[0] is not None:
                ticks['min'] = limits[0]
            if limits[1] is not None:
                ticks['max'] = limits[1]
            if limits[2] is not None:
                ticks['stepSize'] = limits[2]

    def _make_callback(self, data, max_data=100):
        '''
        Create a data callback function for the given object

        Parameters
        ----------
        data : Window or DataFrame-like or callable
            The object to retrieve data from
        max_data : int, optional
            The maximum number of rows of data to retrieve

        Returns
        -------
        function

        '''
        from .windows import BaseWindow
        from .utils.events import get_dataframe

        if isinstance(data, BaseWindow):

            lock = threading.RLock()
            empty_df = get_dataframe(data)
            state = dict(df=empty_df, reset=False, updated=False)

            def on_event(sock, event):
                with lock:
                    if state['reset']:
                        state['df'] = event.tail(self.max_data)
                    else:
                        state['df'] = pd.concat([state['df'], event],
                                                **CONCAT_OPTIONS).tail(self.max_data)
                    state['updated'] = True

            sub = data.create_subscriber(mode='streaming',
                                         interval=self.interval,
                                         pagesize=self.max_data,
                                         on_event=on_event)

            def data_callback(initial=False, max_data=max_data, terminate=False):
                if initial:
                    sub.start()
                if terminate:
                    sub.stop()
                    return
                if state['updated']:
                    with lock:
                        out = state['df'].tail(max_data)
                        state['reset'] = True
                        state['updated'] = False
                        return out
                return empty_df

        elif hasattr(data, 'tail'):
            def data_callback(initial=False, max_data=max_data, terminate=False):
                return data.tail(max_data)

        else:
            def data_callback(initial=False, max_data=max_data, terminate=False):
                return data(initial=initial, max_data=max_data,
                            terminate=terminate).tail(max_data)

        return data_callback

    def _get_data(self, data, initial=False, terminate=False):
        '''
        Retrieve new data points

        Parameters
        ----------
        data : dict
            The data definition
        initial : bool, optional
            Is this the first time the callback is being executed?
        terminate : bool, optional
            Should allocated resources be cleaned up?

        Returns
        -------
        ([labels], [displayed-labels], [datasets-dicts])

        '''
        df = self._data_callback(initial=initial, terminate=terminate)

        if terminate:
            return [], []

        if self._var_generator:
            extra = self._var_generator(df)
        else:
            extra = {}

        df = df.reset_index()
        labels = []
        displayed_labels = None
        datasets = []
        num_obs = 0

        for i, dataset in enumerate(data['datasets']):
            key = dataset['_key']
            x = dataset['_x']
            y = dataset['_y']
            x_labels = dataset.get('_displayed_labels', None)
            radius = dataset['_radius']
            ctype = dataset['type']
            use_labels = None

            if initial or re.match(r'^\s*var\s*\([^\)]+\)\s*$', dataset.get('label', '')):
                out = copy.deepcopy(dataset)
            else:
                out = dict(_key=key)

            if ctype in ['scatter', 'area', 'line']:
                out['data'] = self._get_xy_data(df, x, y, gen_vars=extra)
                use_labels = out['data']
                num_obs = len(out['data'])

            elif ctype in ['bubble']:
                out['data'] = self._get_xyz_data(df, x, y, radius, gen_vars=extra)
                use_labels = out['data']
                num_obs = len(out['data'])

            else:
                labels, displayed_labels, obs = \
                    self._get_labeled_data(df, x, y,
                                           displayed_labels=x_labels, gen_vars=extra)
                out['data'] = obs
                num_obs = len(out['data'])

            self._set_dynamic_vars(dataset, out, df, extra)

            datasets.append(out)

        if not labels:
            if use_labels:
                labels = [item['x'] for item in use_labels]
            else:
                labels = [None] * num_obs

        return self._normalize_data(labels, displayed_labels, datasets, initial=initial)

    def _alias_colors(self, dataset, color):

        def autoify(value):
            if isinstance(value, (tuple, list)):
                return ['auto(%s)' % x for x in value]
            return 'auto(%s)' % value

        if not dataset.get('borderColor') and not dataset.get('backgroundColor'):
            dataset['borderColor'] = autoify(color)
            dataset['backgroundColor'] = autoify(color)

        elif dataset.get('borderColor') and not dataset.get('backgroundColor'):
            dataset['backgroundColor'] = autoify(dataset['borderColor'])

        elif dataset.get('backgroundColor') and not dataset.get('borderColor'):
            if isinstance(dataset['backgroundColor'], six.string_types):
                if dataset['backgroundColor'].startswith('url'):
                    dataset['borderColor'] = autoify(color)
                else:
                    dataset['borderColor'] = autoify(dataset['backgroundColor'])

    def _normalize_data(self, labels, displayed_labels, datasets, initial=False):
        out = []
        palette = self.palette(len(datasets))

        for i, dataset in enumerate(datasets):

            if 'label' not in dataset or not isinstance(dataset['label'], list):
                if initial:
                    self._alias_colors(dataset, palette[i])
                out.append(dataset)
                continue

            finished = set()

            for i, label in enumerate(dataset['label']):
                if label in finished:
                    continue
                else:
                    finished.add(label)

                ds = copy.deepcopy(dataset)

                self._apply_mask(label, ds['label'], ds['data'])

                if isinstance(ds.get('borderColor'), list):
                    ds['borderColor'] = ds['borderColor'][i]
                if isinstance(ds.get('backgroundColor'), list):
                    ds['backgroundColor'] = ds['backgroundColor'][i]
                if isinstance(ds.get('pointStyle'), list):
                    ds['pointStyle'] = ds['pointStyle'][i]
                if ds.get('pointStyle'):
                    ds['pointStyle'] = camelize(ds['pointStyle'])

                ds['label'] = label
                ds['_key'] = '%s-%s' % (ds['_key'], ds['label'])

                self._alias_colors(ds, palette[i % len(palette)])

                out.append(ds)

        return labels, displayed_labels, out

    def _apply_mask(self, label, labels, value, repl=None):
        for i, item in enumerate(labels):
            if item != label:
                value[i] = repl

    def _set_dynamic_vars(self, dataset, out, *data):
        '''
        Set the values of attributes set by variables

        Parameters
        ----------
        dataset : dict
            The input chart dataset configuration
        out : dict
            The output chart dataset configuration
        *data : one-or-more DataFrame-or-dicts
            The data sources for any variables

        '''
        for name in ['label', 'borderColor', 'backgroundColor', 'pointStyle']:
            varname = self._extract_varname(dataset.get(name, ''))
            if varname is not None:
                values = self._get_values(varname, *data)
                if values is not None:
                    if name == 'pointStyle':
                        out[name] = [camelize(x) for x in values]
                    else:
                        out[name] = values

    def _extract_varname(self, value):
        '''
        Extract variable name from parameter value

        Parameter values that are indicated by variables must have the
        form ``var(name)``.

        Parameters
        ----------
        value : string
            The string to look for a variable name

        Returns
        -------
        string
            The name of the variable, if one is found
        None
            If no variable name is found

        '''
        if isinstance(value, six.string_types):
            m = re.match(r'^\s*var\s*\(\s*([^\)]+)\s*\)\s*$', value)
            if m:
                return m.group(1)

    def _get_values(self, name, *data):
        '''
        Retrieve the list of values for the given data name

        Parameters
        ----------
        name : string
            The name of the variable to look for
        *data : one-or-more DataFrames-or-dicts
            The data sets to look in

        Returns
        -------
        list-of-strings
            If the name was found in one of the data sets
        None
            If the name was not found in any of the data sets

        '''
        for item in data:
            if name in item:
                return ['%s' % value for value in item[name]]

    def _add_dataset(self, params, x, y, radius=None):
        '''
        Add a data set to the chart configuration

        Parameters
        ----------
        params : dict
            The chart parameters
        x : string
            The independent variable name
        y : string
            The dependent variable name
        radius : string, optional
            The variable name of the variable that indicates the radius
            of the points in a bubble chart

        '''
        if not self.data['datasets'] or params['type'] in ['pie', 'bar', 'horizontalBar']:
            self.type = params['type']
        params['_x'] = x
        params['_y'] = y
        params['_radius'] = radius
        params['_key'] = 'ds%02d' % len(self.data['datasets'])
        params['data'] = []
        self.data['datasets'].append(params)

    def _get_xy_data(self, data, x, y, gen_vars=None):
        '''
        Return the requested data in [{'x': ..., 'y': ...}] form

        Parameters
        ----------
        data : DataFrame
            The DataFram to retrieve data from
        x : string
            The variable for the x-axis
        y : string
            The variable for the y-axis

        Returns
        -------
        list-of-dicts

        '''
        gen_vars = gen_vars or {}
        out = {}
        out_cols = [x, y]
        if x in gen_vars:
            out[x] = gen_vars[x]
            out_cols[0] = None
        if y in gen_vars:
            out[y] = gen_vars[y]
            out_cols[1] = None
        out_cols = [col for col in out_cols if col is not None]
        if out_cols:
            out.update(data.loc[:, out_cols].to_dict(orient='list'))
        return [dict(x=x, y=y) for x, y in zip(out[x], out[y])]

    def _get_xyz_data(self, data, x, y, r, gen_vars=None):
        '''
        Return the requested data in [{'x': ..., 'y': ..., 'r': ...}] form

        Parameters
        ----------
        data : DataFrame
            The DataFram to retrieve data from
        x : string
            The variable for the x-axis
        y : string
            The variable for the y-axis
        r : string
            The variable for the radius of the points

        Returns
        -------
        list-of-dicts

        '''
        gen_vars = gen_vars or {}
        out = {}
        out_cols = [x, y, r]
        if x in gen_vars:
            out[x] = gen_vars[x]
            out_cols[0] = None
        if y in gen_vars:
            out[y] = gen_vars[y]
            out_cols[1] = None
        if r in gen_vars:
            out[r] = gen_vars[r]
            out_cols[2] = None
        out_cols = [col for col in out_cols if col is not None]
        if out_cols:
            out.update(data.loc[:, out_cols].to_dict(orient='list'))
        return [dict(x=x, y=y, r=r) for x, y, r in zip(out[x], out[y], out[r])]

    def _get_labeled_data(self, data, x, y, displayed_labels=None, gen_vars=None):
        '''
        Return the requested data as a tuple of lables and data points

        Parameters
        ----------
        data : DataFrame
            The DataFrame to retrieve data from
        x : string
            The variable for the x-axis
        y : string
            The variable for the y-axis
        displayed_labels : string
            The variable containing x-axis labels

        Returns
        -------
        ([labels], [displayed-labels], [data-points])

        '''
        gen_vars = gen_vars or {}
        out = {}
        out_cols = [x, displayed_labels, y]
        if x in gen_vars:
            out[x] = gen_vars[x]
            out_cols[0] = None
        if displayed_labels in gen_vars:
            out[displayed_labels] = gen_vars[displayed_labels]
            out_cols[1] = None
        if y in gen_vars:
            out[y] = gen_vars[y]
            out_cols[2] = None
        out_cols = [col for col in out_cols if col is not None]
        if out_cols:
            out.update(data.loc[:, out_cols].to_dict(orient='list'))
        if displayed_labels:
            return out[x], out[displayed_labels], out[y]
        return out[x], None, out[y]

    def pie(self, x, y, line_color=None, line_width=1,
            fill_color=None, rotation=None, circumference=None):
        '''
        Create a pie chart

        Parameters
        ----------
        x : string
            The slice categories
        y : string
            The value of the slice
        line_color : string, optional
            CSS color of the arc borders
        line_width : string, optional
            The width of the arc borders
        fill_color : string, optional
            CSS fill color for the arcs
        rotation : float, optional
            The angle, in radians, to start drawing arcs from
        circumference : float, optional
            The angle, in radians, to sweep the arcs

        '''
        cutout_percentage = 0
        return self.donut(**{k: v for k, v in locals().items() if k != 'self'})

    def donut(self, x, y, line_color=None, line_width=1,
              fill_color=None, cutout_percentage=None, rotation=None,
              circumference=None, color=None):
        '''
        Create a donut chart

        Parameters
        ----------
        x : string
            The slice categories
        y : string
            The value of the slice
        line_color : string, optional
            CSS color of the arc borders
        line_width : string, optional
            The width of the arc borders
        fill_color : string, optional
            CSS fill color for the arcs
        cutout_percentage : int, optional
            The percentage of the center of the chart that is empty
        rotation : float, optional
            The angle, in radians, to start drawing arcs from
        circumference : float, optional
            The angle, in radians, to sweep the arcs

        '''
        donut_param_map = dict(line_color='borderColor',
                               line_width='borderWidth', color='*borderColor',
                               fill_color='backgroundColor',
                               cutout_percentage='cutoutPercentage',
                               rotation='rotation', circumference='circumference')

        params = {donut_param_map[k]: v for k, v in locals().items()
                  if k in donut_param_map and v is not None}

        if '*borderColor' in params and 'borderColor' not in params:
            params['borderColor'] = params.pop('*borderColor')

        if params.get('borderWidth', 0) == 0:
            params['borderWidth'] = 0.001

        if params.get('borderColor') is None:
            params['borderColor'] = self.palette(12)

        params['type'] = 'doughnut'

        # These affect all donut charts on this axis
        if cutout_percentage is not None:
            self.options['cutoutPercentage'] = cutout_percentage
        if rotation is not None:
            self.options['rotation'] = rotation
        if circumference is not None:
            self.options['circumference'] = circumference

        self._add_dataset(params, x, y)

    doughnut = donut

    def area(self, x, y, label=None, line_color=None,
             line_width=None, line_dash=None, line_dash_offset=None,
             line_cap_style=None, line_join_style=None, line_tension=None,
             color=None, fill_color=None, fill_mode='origin',
             point_radius=0, point_style=None, display=True,
             span_gaps=False, stacked=False):
        '''
        Create an area chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        label : string, optional
            The label for the data set (used in the legend)
        line_color : string, optional
            CSS color value for the stroke color
        line_width : int or float, optional
            The stroke thickness
        line_dash : list-of-ints or list-of-floats, optional
            The line dash pattern
        line_dash_offset : int or float, optional
            The offset to start the line dash pattern
        line_cap_style : string, optional
            The method of drawing line endpoints: 'butt', 'round', 'square'
        line_join_style : string, optional
            The method of drawing line intersections: 'bevel', 'round', 'miter'
        line_tension : int or float, optional
            Bezier curve tension of the line
        color : string, optional
            Alias of line_color
        fill_color : string, optional
            The fill color under the line
        fill_mode : string or int or bool, optional
            Determines the method of filling.
                * int - fill to the dataset at the given index
                * numeric string (e.g., '-1', '-2', '+1') - fill to
                  relative dataset index
                * string - fill to 'start', 'end', 'origin'
                * bool - True means fill to origin, False means no fill
        point_radius : int or float, optional
            The radius of the data points
        display : bool, optional
            Should the line itself be displayed?
        span_gaps : bool, optional
            Should missing values be spanned?

        '''
        return self.line(**{k: v for k, v in locals().items() if k != 'self'})

    def line(self, x, y, label=None, line_color=None,
             line_width=None, line_dash=None, line_dash_offset=None,
             line_cap_style=None, line_join_style=None, line_tension=None,
             color=None, fill_color=None, fill_mode=False,
             point_radius=0, point_style=None, display=True,
             span_gaps=False, stacked=False):
        '''
        Create a line chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        label : string, optional
            The label for the data set (used in the legend)
        line_color : string, optional
            CSS color value for the stroke color
        line_width : int or float, optional
            The stroke thickness
        line_dash : list-of-ints or list-of-floats, optional
            The line dash pattern
        line_dash_offset : int or float, optional
            The offset to start the line dash pattern
        line_cap_style : string, optional
            The method of drawing line endpoints: 'butt', 'round', 'square'
        line_join_style : string, optional
            The method of drawing line intersections: 'bevel', 'round', 'miter'
        line_tension : int or float, optional
            Bezier curve tension of the line
        color : string, optional
            Alias of line_color
        fill_color : string, optional
            The fill color under the line
        fill_mode : string or int or bool, optional
            Determines the method of filling.
                * int - fill to the dataset at the given index
                * numeric string (e.g., '-1', '-2', '+1') - fill to
                  relative dataset index
                * string - fill to 'start', 'end', 'origin'
                * bool - True means fill to origin, False means no fill
        point_radius : int or float, optional
            The radius of the data points
        display : bool, optional
            Should the line itself be displayed?
        span_gaps : bool, optional
            Should missing values be spanned?

        '''

        line_param_map = dict(label='label', line_color='borderColor',
                              line_width='borderWidth', line_dash='borderDash',
                              line_dash_offset='borderDashOffset',
                              line_cap_style='borderCapStyle',
                              line_join_style='borderJoinStyle',
                              line_tension='lineTension',
                              color='*borderColor', fill_color='backgroundColor',
                              fill_mode='fill', point_radius='pointRadius',
                              point_style='pointStyle', display='showLine',
                              span_gaps='spanGaps')

        params = {line_param_map[k]: v for k, v in locals().items()
                  if k in line_param_map and v is not None}

        if '*borderColor' in params and 'borderColor' not in params:
            params['borderColor'] = params.pop('*borderColor')

        if params.get('pointStyle'):
            params['pointStyle'] = camelize(params['pointStyle'])
        if params.get('borderWidth', 1) == 0:
            params['borderWidth'] = 0.001

        params['type'] = 'line'

        if 'label' not in params:
            params['label'] = y

        if stacked:
            self.options['scales']['yAxes'][0]['stacked'] = True

        self._add_dataset(params, x, y)

    def scatter(self, x, y, label=None, color=None, fill_color=None,
                line_color=None, line_width=None, point_radius=3,
                point_style=None, display=False):
        '''
        Create a scatter chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        label : string, optional
            The label for the data set (used in the legend)
        color : string, optional
            CSS color value for the stroke color
        fill_color : string, optional
            The fill color for data points
        line_color : string, optional
            Alias for color
        line_width : int or float, optional
            The stroke thickness of the data points
        point_radius : int or float, optional
            The radius of the data points
        point_style : string, optional
            The point style for each data point: 'circle', 'cross', 'cross-rot',
            'dash', 'line', 'rect', 'rect-rounded', 'star', 'triangle'
        display : bool, optional
            Should the line itself be displayed?

        '''
        scatter_param_map = dict(label='label',
                                 color='*borderColor', line_color='borderColor',
                                 fill_color='backgroundColor',
                                 line_width='borderWidth',
                                 point_radius='pointRadius',
                                 point_style='pointStyle', display='showLine')

        params = {scatter_param_map[k]: v for k, v in locals().items()
                  if k in scatter_param_map and v is not None}

        if '*borderColor' in params and 'borderColor' not in params:
            params['borderColor'] = params.pop('*borderColor')

        if params.get('pointStyle'):
            params['pointStyle'] = camelize(params['pointStyle'])

        if params.get('borderWidth', 1) == 0:
            params['borderWidth'] = 0.001

        params['type'] = 'scatter'

        if 'label' not in params:
            params['label'] = y

        self._add_dataset(params, x, y)

    def bubble(self, x, y, radius, label=None, color=None, fill_color=None,
               line_color=None, line_width=None, point_style=None, display=False):
        '''
        Create a bubble chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        radius : string
            The variable name that contains the point radii
        label : string, optional
            The label for the data set (used in the legend)
        color : string, optional
            CSS color value for the stroke color
        fill_color : string, optional
            The fill color for data points
        line_color : string, optional
            Alias for color
        line_width : int or float, optional
            The stroke thickness of the data points
        point_radius : int or float, optional
            The radius of the data points
        point_style : string, optional
            The point style for each data point: 'circle', 'cross', 'cross-rot',
            'dash', 'line', 'rect', 'rect-rounded', 'star', 'triangle'
        display : bool, optional
            Should the line itself be displayed?

        '''

        scatter_param_map = dict(label='label', color='*borderColor',
                                 line_color='borderColor',
                                 fill_color='backgroundColor',
                                 line_width='borderWidth',
                                 point_style='pointStyle', display='showLine')

        params = {scatter_param_map[k]: v for k, v in locals().items()
                  if k in scatter_param_map and v is not None}

        if '*borderColor' in params and 'borderColor' not in params:
            params['borderColor'] = params.pop('*borderColor')

        if params.get('pointStyle'):
            params['pointStyle'] = camelize(params['pointStyle'])

        if params.get('borderWidth', 1) == 0:
            params['borderWidth'] = 0.001

        params['type'] = 'bubble'

        if 'label' not in params:
            params['label'] = y

        self._add_dataset(params, x, y, radius)

    def bar(self, x, y, label=None, displayed_labels=None,
            line_color=None, line_width=1, color=None,
            fill_color=None, border_skipped=None, bar_percentage=None,
            category_percentage=None, bar_thickness=None, max_bar_thickness=None,
            stack_id=None, orientation='vertical'):
        '''
        Create a bar chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        label : string, optional
            The label for the data set (used in the legend)
        displayed_labels : string
            The variable name containing labels to display in place of x values
        line_color : string, optional
            CSS color value for the stroke color
        line_width : int or float, optional
            The stroke thickness of the data points
        color : string, optional
            Alias for color
        fill_color : string, optional
            The fill color for data points
        border_skipped : string, optional
            Border to avoid drawing: 'bottom', 'top', 'left', 'right'
        bar_percentage : float, optional
            The percentage of space within the category to use
        category_percentage : float, optional
            The percentage of space to use for each category.
            ``bar_percentage`` is the percentage within this space to use
            for each bar.
        bar_thickness : int or float, optional
            The manually-set thickness of the bars
        stack_id : string, optional
            The stack identifier for the bar.  All bars using the same stack
            identifier are stacked onto each other.
        orientation : string, optional
            The orientation of the chart: 'horizontal', 'vertical'

        '''
        bar_param_map = dict(label='label', line_color='borderColor',
                             line_width='borderWidth', color='*borderColor',
                             fill_color='backgroundColor',
                             border_skipped='borderSkipped',
                             stack_id='stack', displayed_labels='_displayed_labels')

        params = {bar_param_map[k]: v for k, v in locals().items()
                  if k in bar_param_map and v is not None}

        if '*borderColor' in params and 'borderColor' not in params:
            params['borderColor'] = params.pop('*borderColor')

        horizontal = orientation.startswith('horiz')

        params['type'] = horizontal and 'horizontalBar' or 'bar'

        if params.get('borderWidth', 1) == 0:
            params['borderWidth'] = 0.001

        if 'label' not in params:
            params['label'] = y

        self._add_dataset(params, x, y)

        # These affect all bar charts on this axis
        if bar_percentage is not None:
            self.options['scales']['xAxes'][-1]['barPercentage'] = bar_percentage
            self.options['scales']['yAxes'][-1]['barPercentage'] = bar_percentage
        if category_percentage is not None:
            self.options['scales']['xAxes'][-1]['categoryPercentage'] = category_percentage
            self.options['scales']['yAxes'][-1]['categoryPercentage'] = category_percentage
        if bar_thickness is not None:
            self.options['scales']['xAxes'][-1]['barThickness'] = bar_thickness
            self.options['scales']['yAxes'][-1]['barThickness'] = bar_thickness
        if max_bar_thickness is not None:
            self.options['scales']['xAxes'][-1]['maxBarThickness'] = max_bar_thickness
            self.options['scales']['yAxes'][-1]['maxBarThickness'] = max_bar_thickness

    def hbar(self, x, y, **kwargs):
        '''
        Create a horizontal bar chart

        Parameters
        ----------
        x : string
            The x-axis variable name
        y : string
            The y-axis variable name
        label : string, optional
            The label for the data set (used in the legend)
        line_color : string, optional
            CSS color value for the stroke color
        line_width : int or float, optional
            The stroke thickness of the data points
        color : string, optional
            Alias for color
        fill_color : string, optional
            The fill color for data points
        border_skipped : string, optional
            Border to avoid drawing: 'bottom', 'top', 'left', 'right'
        bar_percentage : float, optional
            The percentage of space within the category to use
        category_percentage : float, optional
            The percentage of space to use for each category.
            ``bar_percentage`` is the percentage within this space to use
            for each bar.
        bar_thickness : int or float, optional
            The manually-set thickness of the bars
        stack_id : string, optional
            The stack identifier for the bar.  All bars using the same stack
            identifier are stacked onto each other.

        '''
        return self.bar(x, y, orientation='horizontal', **kwargs)

    def _repr_html_(self, plot_width=None, plot_height=None):
        plot_id = 'chart_%s' % str(uuid.uuid4()).replace('-', '_')
        plot_width = plot_width or self.plot_width
        plot_height = plot_height or self.plot_height
        max_data = self.max_data
        steps = self.steps

        loc_data = copy.deepcopy(self.data)
        loc_options = copy.deepcopy(self.options)

        # Use custom label generator for everything except donut/pie
        if self.type not in ['pie', 'donut']:
            loc_options['legend']['labels']['generateLabels'] = 'function(generateLabels)'

        # Set initial data
        data = {}
        data['labels'], data['displayed_labels'], data['datasets'] = \
            self._get_data(loc_data, initial=True)

        if steps <= 0:
            out = '''
                <canvas id='%(plot_id)s_canvas'
                        style='width: %(plot_width)spx;
                               height: %(plot_height)spx'></canvas>

                <script language='javascript'>
                <!--
                requirejs.config({
                    paths: {
                        Chart: ['//cdnjs.cloudflare.com/ajax/libs/Chart.js/2.7.0/Chart.min']
                    }
                });

                require(['jquery', 'Chart'], function($, Chart) {

                    var generateLabels = function( chart ) {
                        var data = chart.data;
                        var labels = [];
                        var ds_labels = [];
                        var prev_key = '';

                        if ( !Array.isArray(data.datasets) ) {
                            return labels;
                        }

                        for ( var i=0; i < data.datasets.length; i++ ) {
                            var dataset = data.datasets[i];
                            var entry = {
                                text: dataset.label,
                                fillStyle: (!Array.isArray(dataset.backgroundColor) ?
                                            dataset.backgroundColor :
                                            dataset.backgroundColor[0]),
                                hidden: !chart.isDatasetVisible(i),
                                lineCap: dataset.borderCapStyle,
                                lineDash: dataset.borderDash,
                                lineDashOffset: dataset.borderDashOffset,
                                lineJoin: dataset.borderJoinStyle,
                                lineWidth: dataset.borderWidth,
                                strokeStyle: dataset.borderColor,
                                pointStyle: dataset.pointStyle,

                                // Below is extra data used for toggling the datasets
                                datasetIndex: i
                            }

                            if ( dataset._key &&
                                 dataset._key.split('-')[0] === prev_key ) {
                                ds_labels.push(entry);
                            }
                            else {
                                ds_labels.sort(function(a, b) {
                                    if ( a.text < b.text ) return -1;
                                    if ( a.text > b.text ) return 1;
                                    return 0;
                                });
                                for ( var j=0; j < ds_labels.length; j++ ) {
                                    labels.push(ds_labels[j]);
                                }
                                prev_key = dataset._key.split('-')[0];
                                ds_labels = [entry];
                            }
                        }

                        ds_labels.sort(function(a, b) {
                            if ( a.text < b.text ) return -1;
                            if ( a.text > b.text ) return 1;
                            return 0;
                        });
                        for ( var j=0; j < ds_labels.length; j++ ) {
                            labels.push(ds_labels[j]);
                        }

                        return labels;
                    };

                    var auto_colors = function ( data ) {
                        var ctx = document.getElementById('%(plot_id)s_canvas')
                                          .getContext('2d');
                        var datasets = data.datasets;
                        var Color = Chart.helpers.color;
                        if ( !data.datasets || data.datasets.length == 0 ) {
                            return data;
                        }
                        for ( var i=0; i < datasets.length; i++ ) {
                            var ds = datasets[i];
                            if ( ds.backgroundColor ) {
                                if ( Array.isArray(ds.backgroundColor) ) {
                                    for ( var j=0; j < ds.backgroundColor.length; j++ ) {
                                        var bgcolor = ds.backgroundColor[j];
                                        if ( bgcolor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                            bgcolor = bgcolor.replace(/^\s*auto\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                            ds.backgroundColor[j] = Color(bgcolor).clearer(0.6).rgbString();
                                        }
                                        else if ( bgcolor.match(/^\s*url\s*\(.+\)\s*$/) ) {
                                            var img = new Image();
                                            img.src = bgcolor.replace(/^\s*url\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                            img.onload = function () {
                                                ds.backgroundColor[j] = ctx.createPattern(img, 'repeat');
                                            };
                                        }
                                    }
                                }
                                else if ( ds.backgroundColor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                    var bgcolor = ds.backgroundColor.replace(/^\s*auto\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                    ds.backgroundColor = Color(bgcolor).clearer(0.6).rgbString();
                                }
                                else if ( ds.backgroundColor.match(/^\s*url\s*\(.+\)\s*$/) ) {
                                    var img = new Image();
                                    img.src = ds.backgroundColor.replace(/^\s*url\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                    img.onload = function () {
                                        ds.backgroundColor = ctx.createPattern(img, 'repeat');
                                    };
                                }
                            }
                            if ( ds.borderColor ) {
                                if ( Array.isArray(ds.borderColor) ) {
                                    for ( var j=0; j < ds.borderColor.length; j++ ) {
                                        var fgcolor = ds.borderColor[j];
                                        if ( fgcolor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                            fgcolor = fg.replace(/^\s*auto\s*\(\s*(.*)\s*\)\s*$/, '$1');
                                            ds.borderColor[j] = Color(fgcolor).opaquer(1).rgbString();
                                        }
                                    }
                                } else if ( ds.borderColor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                    var fgcolor = ds.borderColor.replace(/^\s*auto\s*\(\s*(.*)\s*\)\s*$/, '$1');
                                    ds.borderColor = Color(fgcolor).opaquer(1).rgbString();
                                }
                            }
                        }
                        return data;
                    };

                    var ctx = document.getElementById('%(plot_id)s_canvas').getContext('2d');
                    var chart = new Chart(ctx, {
                        type: '%(chart_type)s',
                        data: auto_colors(%(data)s),
                        options: %(options)s
                    });
                });
                //-->
                </script>
            ''' % dict(plot_id=plot_id, plot_width=plot_width, plot_height=plot_height,
                       chart_type=self.type, data=dump_json(data),
                       options=dump_json(loc_options))

            return re.sub(r'([\'"])function\((\w+)\)\1', r'\2', out)

        # Comm setup
        state = dict(paused=False, kill=False)

        def target_func(comm, msg): # pragma: no cover
            ''' Setup comm object '''
            @comm.on_msg
            def on_msg(msg):
                ''' Handle comm messages '''
                data = msg['content']['data']
                if 'command' in data:
                    command = data['command']
                    if command == 'stop':
                        state['kill'] = True
                    elif command == 'pause':
                        state['paused'] = True
                    elif command in ['start', 'play']:
                        state['paused'] = False

            @comm.on_close
            def on_close(msg):
                ''' Close the comm '''
                state['kill'] = True

            def do_plot(_get_data, loc_data, steps, interval):
                ''' Loop and update the data as needed '''
                while steps > 0 and not state['kill']:

                    if not state['paused']:
                        labels, displayed_labels, datasets = _get_data(loc_data)
                        if datasets and datasets[0]:
                            comm.send({'labels': labels,
                                       'displayed_labels': displayed_labels,
                                       'datasets': datasets})

                    time.sleep(interval / 1000.)
                    steps -= 1

                comm.send({'command': 'stop'})

                _get_data(loc_data, terminate=True)

            def kill():
                ''' Indicate that the image drawing loop should be killed '''
                state['kill'] = True

            weakref.ref(self, kill)

            threading.Thread(target=do_plot, name=plot_id,
                             args=(self._get_data, loc_data, self.steps,
                                   self.interval)).start()

        try:
            get_ipython().kernel.comm_manager.register_target(plot_id, target_func)
        except NameError:
            warnings.warn('Streaming figures are only supported in Jupyter notebooks',
                          RuntimeWarning) 

        out = '''
            <canvas id='%(plot_id)s_canvas'
                    style='width: %(plot_width)spx; height: %(plot_height)spx'></canvas>

            <style type="text/css">
            .stream-plot-controls {
                border: 1px solid #e0e0e0;
                border-radius: 5px;
                padding: 8px;
                display: inline-block;
            }
            .stream-plot-control {
                color: #909090;
                font-size: 150%%;
                margin-right: 1em;
                margin-left: 1em;
            }
            .fa-play.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #00c000;
                color: #00c000;
            }
            .fa-pause.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #fdeb61;
                color: #d37a2d;
            }
            .fa-stop.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #ff0000;
                color: #ff0000;
            }
            .fa-play.stream-plot-control.enabled { color: #00c000; }
            .fa-pause.stream-plot-control.enabled { color: #d37a2d; }
            .fa-stop.stream-plot-control.enabled { color: #ff0000; }
            .stream-plot-control.disabled { color: #d0d0d0; }
            </style>

            <div id="streaming-plot-%(plot_id)s"
                 style="text-align: center; padding-top:10px; width: %(plot_width)spx">
            <div class="stream-plot-controls" id="stream-plot-%(plot_id)s">
              <i class="fa fa-play enabled stream-plot-control"
                 title="Stream data into figure" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-play"></i>
              <i class="fa fa-pause stream-plot-control"
                 title="Pause data updates" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-pause"></i>
              <i class="fa fa-stop stream-plot-control"
                 title="Shutdown data stream" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-stop"></i>
            </div>
            </div>

            <script language='javascript'>
            <!--
            requirejs.config({
                paths: {
                    Chart: ['//cdnjs.cloudflare.com/ajax/libs/Chart.js/2.7.0/Chart.min']
                }
            });

            require(['jquery', 'Chart'], function($, Chart) {

                var afterBuildTicks = function ( scale ) {
                    return;
                    scale.ticks.shift();
                    scale.ticks.pop();
                };

                var generateLabels = function( chart ) {
                    var data = chart.data;
                    var labels = [];
                    var ds_labels = [];
                    var prev_key = '';

                    if ( !Array.isArray(data.datasets) ) {
                        return labels;
                    }

                    for ( var i=0; i < data.datasets.length; i++ ) {
                        var dataset = data.datasets[i];
                        var entry = {
                            text: dataset.label,
                            fillStyle: (!Array.isArray(dataset.backgroundColor) ?
                                        dataset.backgroundColor :
                                        dataset.backgroundColor[0]),
                            hidden: !chart.isDatasetVisible(i),
                            lineCap: dataset.borderCapStyle,
                            lineDash: dataset.borderDash,
                            lineDashOffset: dataset.borderDashOffset,
                            lineJoin: dataset.borderJoinStyle,
                            lineWidth: dataset.borderWidth,
                            strokeStyle: dataset.borderColor,
                            pointStyle: dataset.pointStyle,

                            // Below is extra data used for toggling the datasets
                            datasetIndex: i
                        }

                        if ( dataset._key && dataset._key.split('-')[0] === prev_key ) {
                            ds_labels.push(entry);
                        }
                        else {
                            ds_labels.sort(function(a, b) {
                                if ( a.text < b.text ) return -1;
                                if ( a.text > b.text ) return 1;
                                return 0;
                            });
                            for ( var j=0; j < ds_labels.length; j++ ) {
                                labels.push(ds_labels[j]);
                            }
                            prev_key = dataset._key.split('-')[0];
                            ds_labels = [entry];
                        }
                    }

                    ds_labels.sort(function(a, b) {
                        if ( a.text < b.text ) return -1;
                        if ( a.text > b.text ) return 1;
                        return 0;
                    });
                    for ( var j=0; j < ds_labels.length; j++ ) {
                        labels.push(ds_labels[j]);
                    }

                    return labels;
                };

                var auto_colors = function ( data ) {
                    var ctx = document.getElementById('%(plot_id)s_canvas').getContext('2d');
                    var datasets = data.datasets;
                    var Color = Chart.helpers.color;
                    if ( !data.datasets || data.datasets.length == 0 ) {
                        return data;
                    }
                    for ( var i=0; i < datasets.length; i++ ) {
                        var ds = datasets[i];
                        if ( ds.backgroundColor ) {
                            if ( Array.isArray(ds.backgroundColor) ) {
                                for ( var j=0; j < ds.backgroundColor.length; j++ ) {
                                    var bgcolor = ds.backgroundColor[j];
                                    if ( bgcolor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                        bgcolor = bgcolor.replace(/^\s*auto\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                        ds.backgroundColor[j] = Color(bgcolor).clearer(0.6).rgbString();
                                    }
                                    else if ( bgcolor.match(/^\s*url\s*\(.+\)\s*$/) ) {
                                        var img = new Image();
                                        img.src = bgcolor.replace(/^\s*url\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                        img.onload = function () {
                                            ds.backgroundColor[j] = ctx.createPattern(img, 'repeat');
                                        };
                                    }
                                }
                            }
                            else if ( ds.backgroundColor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                var bgcolor = ds.backgroundColor.replace(/^\s*auto\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                ds.backgroundColor = Color(bgcolor).clearer(0.6).rgbString();
                            }
                            else if ( ds.backgroundColor.match(/^\s*url\s*\(.+\)\s*$/) ) {
                                var img = new Image();
                                img.src = ds.backgroundColor.replace(/^\s*url\s*\(\s*(.+)\s*\)\s*$/, '$1');
                                img.onload = function () {
                                    ds.backgroundColor = ctx.createPattern(img, 'repeat');
                                };
                            }
                        }
                        if ( ds.borderColor ) {
                            if ( Array.isArray(ds.borderColor) ) {
                                for ( var j=0; j < ds.borderColor.length; j++ ) {
                                    var fgcolor = ds.borderColor[j];
                                    if ( fgcolor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                        fgcolor = fg.replace(/^\s*auto\s*\(\s*(.*)\s*\)\s*$/, '$1');
                                        ds.borderColor[j] = Color(fgcolor).opaquer(1).rgbString();
                                    }
                                }
                            } else if ( ds.borderColor.match(/^\s*auto\s*\(.+\)\s*$/) ) {
                                var fgcolor = ds.borderColor.replace(/^\s*auto\s*\(\s*(.*)\s*\)\s*$/, '$1');
                                ds.borderColor = Color(fgcolor).opaquer(1).rgbString();
                            }
                        }
                    }
                    return data;
                };

                var max_data = %(max_data)s;
                var ctx = document.getElementById('%(plot_id)s_canvas').getContext('2d');
                var chart = new Chart(ctx, {
                    type: '%(chart_type)s',
                    data: auto_colors(%(data)s),
                    options: %(options)s
                });

                var set_display_labels = function (value, index, values) {
                    if ( chart.data.displayed_labels && chart.data.displayed_labels[index] ) {
                        return chart.data.displayed_labels[index];
                    }
                    return value;
                };

                if ( chart.options.scales && chart.options.scales.xAxes ) {
                    chart.options.scales.xAxes[0].ticks.callback = set_display_labels;
                }

/*
                var dynamic_x_axis = false;
                if ( chart.config.type === 'line' || chart.config.type === 'area' ||
                     chart.config.type === 'scatter' || chart.config.type === 'bubble' ) {
                    if ( chart.options.scales.xAxes[0].ticks.min == null &&
                         chart.options.scales.xAxes[0].ticks.max == null ) {
                        dynamic_x_axis = true;
                    }
                }

                if ( dynamic_x_axis && chart.options.scales && chart.options.scales.xAxes ) {
                    chart.options.scales.xAxes[0].afterBuildTicks = afterBuildTicks;
                }
*/

                var rotate_linear = function (labels, values, new_values) {
                    Array.prototype.push.apply(values, new_values);
                    while ( values.length > max_data ) {
                        values.shift();
                    }
                    while ( values.length < labels.length ) {
                        values.unshift(null);
                    }
                }

                var rotate_category = function (labels, new_labels, values, new_values) {
                    var idx;

                    if ( !new_labels || new_labels.length == 0 ) return;
                    if ( !new_values || new_values.length == 0 ) return;

                    for ( var i=0; i < new_labels.length; i++ ) {
                        idx = labels.indexOf(new_labels[i]);
                        if ( idx === -1 ) {
                            labels.push(new_labels[i]);
                            values.push(new_values[i]);
                        } else {
                            values[idx] = new_values[i];
                        }
                    }

                    while ( values.length > max_data ) {
                        values.shift();
                    }

                    while ( values.length < labels.length ) {
                        values.unshift(null);
                    }
                }

                var comm = Jupyter.notebook.kernel.comm_manager.new_comm('%(plot_id)s')

                comm.on_msg(function(msg) {
                    var data = auto_colors(msg.content.data);
                    var key = null;
                    var old_datasets = {};
                    var new_datasets = {};
                    var updated_datasets = {};

                    if ( data.labels && data.datasets &&
                         data.labels.length > 0 && data.datasets.length > 0 ) {

                        // Categorize updatable and new datasets
                        for ( var i=0; i < chart.data.datasets.length; i++ ) {
                            old_datasets[chart.data.datasets[i]._key] = chart.data.datasets[i];
                        }
                        for ( var i=0; i < data.datasets.length; i++ ) {
                            if ( old_datasets[data.datasets[i]._key] == null ) {
                                new_datasets[data.datasets[i]._key] = data.datasets[i];
                            } else {
                                updated_datasets[data.datasets[i]._key] = data.datasets[i];
                            }
                        }

                        if ( chart.config.type == 'doughnut' ||
                             chart.config.type == 'pie' ||
                             chart.config.type == 'bar' ||
                             chart.config.type == 'horizontalBar' ) {

                            // Update existing datasets
                            for ( key in updated_datasets ) {
                                rotate_category( chart.data.labels,
                                                   data.labels,
                                                   old_datasets[key].data,
                                                   updated_datasets[key].data );
                            }

                            // Add new datasets
                            for ( key in new_datasets ) {
                                chart.data.datasets.push(new_datasets[key]);
                                rotate_category( chart.data.labels,
                                                   data.labels,
                                                   new_datasets[key].data,
                                                   new_datasets[key].data );
                            }

                            chart.data.displayed_labels = data.displayed_labels;

                        } else {

                            // Rotate labels
                            Array.prototype.push.apply(chart.data.labels, data.labels);
                            while ( chart.data.labels.length > max_data ) {
                                chart.data.labels.shift();
                            }

                            // Update existing datasets
                            for ( key in updated_datasets ) {
                                rotate_linear( chart.data.labels,
                                               old_datasets[key].data,
                                               updated_datasets[key].data );
                            }

                            // Add new datasets
                            for ( key in new_datasets ) {
                                chart.data.datasets.push(new_datasets[key]);
                                rotate_linear( chart.data.labels,
                                               new_datasets[key].data,
                                               [] );
                            }

                            // Shift stale datasets
                            var nulls = [];
                            for ( key in old_datasets ) {
                                if ( updated_datasets[key] == null ) {
                                    if ( nulls.length == 0 ) {
                                        for ( var i=0; i < data.labels.length; i++ ) {
                                            nulls.push(null);
                                        }
                                    }
                                    rotate_linear( chart.data.labels,
                                                   old_datasets[key].data,
                                                   nulls );
                                }
                            }

                            // Reset x-axis limits
/*
                            if ( dynamic_x_axis && chart.options.scales && chart.options.scales.xAxes ) {
                                chart.options.scales.xAxes[0].ticks.min = chart.data.labels[Math.max(Math.min(chart.data.labels.length-1, 2), 0)];
                                chart.options.scales.xAxes[0].ticks.max = chart.data.labels[Math.max(chart.data.labels.length-1, 0)];
                            }
*/
                        }

                        if ( chart.config.type == 'line' ) {
                            chart.update(0);
                        } else {
                            chart.update(%(interval)s * 2);
                        }
                    }

                    if ( data.command ) {
                        if ( data.command == 'stop' ) {
                            $('#%(plot_id)s-play, #%(plot_id)s-pause, #%(plot_id)s-stop')
                                .removeClass('enabled')
                                .addClass('disabled')
                                .css('cursor', 'default');
                            $('#streaming-plot-%(plot_id)s').hide().remove();
                        }
                    }

                    // If plot doesn't exist anymore, stop data
                    if ( $('#streaming-plot-%(plot_id)s').length == 0 ) {
                        comm.send({'command': 'stop'});
                    }
                });

                comm.on_close(function() {
                    comm.send({'command': 'stop'});
                });

                // Send message when plot is removed
                $.event.special.destroyed = {
                    remove: function(o) {
                      if (o.handler) {
                        o.handler()
                      }
                    }
                }
                $('#streaming-plot-%(plot_id)s').bind('destroyed', function() {
                    comm.send({'command': 'stop'});
                });

                var blink = function () {
                  $('#%(plot_id)s-pause.enabled').fadeOut(900).fadeIn(900, blink);
                }

                $('#%(plot_id)s-play').click(function() {
                  if ( $('#%(plot_id)s-play').hasClass('disabled') ) return;
                  if ( $('#%(plot_id)s-play').hasClass('enabled') ) return;
                  $('#%(plot_id)s-pause').removeClass('enabled');
                  $('#%(plot_id)s-play').addClass('enabled');
                  comm.send({'command': 'play'});
                });

                $('#%(plot_id)s-pause').click(function() {
                  if ( $('#%(plot_id)s-pause').hasClass('disabled') ) return;
                  if ( $('#%(plot_id)s-pause').hasClass('enabled') ) {
                    $('#%(plot_id)s-play').click();
                    return;
                  }
                  $('#%(plot_id)s-play').removeClass('enabled');
                  $('#%(plot_id)s-pause').addClass('enabled');
                  comm.send({'command': 'pause'});
                  blink();
                });

                $('#%(plot_id)s-stop').click(function() {
                  if ( $('#%(plot_id)s-stop').hasClass('disabled') ) return;
                    $('#%(plot_id)s-play, #%(plot_id)s-pause, #%(plot_id)s-stop')
                    .removeClass('enabled')
                    .addClass('disabled')
                    .css('cursor', 'default');
                  comm.send({'command': 'stop'});
                  $('#streaming-plot-%(plot_id)s').hide().remove();
                });

            });
            //-->
            </script>
        ''' % dict(plot_id=plot_id, steps=steps, interval=self.interval,
                   plot_width=plot_width, plot_height=plot_height,
                   chart_type=self.type, data=dump_json(data),
                   options=dump_json(self.options), max_data=max_data)

        return re.sub(r'([\'"])function\((\w+)\)\1', r'\2', out)


class StreamingImages(object):
    '''
    Streaming Images

    Attributes
    ----------
    data_callback : callable
        The function to call for data updates
    interval : int, optional
        The length of each step in milliseconds
    steps : int, optional
        The maximum number of steps to iterate

    Parameters
    ----------
    data_callback : callable
        The function to call for data updates
    image_key : string, optional
        The key for the image data in the callback result.  If a key value
        is not set, but there is only one key in the data, the data at that
        key will be used.
    steps : int, optional
        The maximum number of steps to iterate
    interval : int, optional
        The length of each step in milliseconds
    size : string or int or tuple, optional
        The size of the displayed image.  It can take the following values:
            * 'contain' - to indicate that the image should be scaled to
              fit inside the plot area
            * 'cover' - to indicate that the image should scale to cover
              the entire plot area
            * int - to indicate a square with a pixel width of height of
              the given integer
            * (int, int) - to indicate width and height values in pixels
    plot_width : int, optional
        The width of the plot in pixels
    plot_height : int, optional
        The height of the plot in pixels
    annotations : Window, optional
        Window containing image highlighting information
    transformers : function or list-of-functions, optional
        Functions to apply to each image prior to displaying

    Returns
    -------
    :class:`StreamingImages`

    '''

    def __init__(self, data_callback, image_key=None, steps=1e5, interval=1000,
                 size=None, plot_width=900, plot_height=400, annotations=None,
                 transformers=None, **kwargs):
        self.data_callback = data_callback
        self._image_key = image_key
        self.plot_width = plot_width
        self.plot_height = plot_height
        self.size = size
        self.steps = steps
        self.interval = interval
        self.annotations = annotations
        self.transformers = listify(transformers)

    def _repr_html_(self, plot_height=None, plot_width=None):
        plot_id = 'image_comm_%s' % str(uuid.uuid4()).replace('-', '_')

        state = dict(paused=False, kill=False, image_key=self._image_key)

        def target_func(comm, msg): # pragma: no cover
            ''' Setup comm object '''
            from PIL import Image

            @comm.on_msg
            def on_msg(msg):
                ''' Handle comm messages '''
                data = msg['content']['data']
                if 'command' in data:
                    command = data['command']
                    if command == 'stop':
                        state['kill'] = True
                    elif command == 'pause':
                        state['paused'] = True
                    elif command in ['start', 'play']:
                        state['paused'] = False

            @comm.on_close
            def on_close(msg):
                ''' Close the comm '''
                state['kill'] = True

            def do_plot(data_callback, steps, interval, annotations, transformers):
                ''' Loop and update the image as needed '''
                from .windows import BaseWindow

                previous = None
                initial = True

                while steps > 0 and not state['kill']:

                    if not state['paused']:
                        data = data_callback(initial=initial, max_data=1)
                        initial = False
                        if state['image_key'] is None and len(data.keys()) == 1:
                            state['image_key'] = list(data.keys())[0]
                        if state['image_key'] in data:
                            data = data[state['image_key']]
                            if len(data):
                                modified = False
                                data = data[-1]
                                if Image.isImageType(data):
                                    img = data
                                else:
                                    img = Image.open(io.BytesIO(data))
                                img_format = img.format

                                if transformers is not None and transformers:
                                    for trans in transformers:
                                        img = trans(img) 
                                        modified = True

                                if annotations is None:
                                    pass
                                elif isinstance(annotations, BaseWindow):
                                    if ('_nObjects_' in annotations.columns and
                                        '_Object0_' in annotations.columns):
                                        img = highlight_image(img, annotations)
                                        modified = True

                                if modified:
                                    data = io.BytesIO()
                                    img.save(data, format=img_format)
                                    data = data.getvalue()

                                data = b'data:image/' + img_format.lower().encode('utf-8') + \
                                       b';base64,' + base64.b64encode(data)
                                if previous is None or previous != data:
                                    random = id(object())
                                    comm.send({'url': data,
                                               'id': random,
                                               'format': img_format,
                                               'length': len(data),
                                               'size': list(img.size)})
                                    previous = data

                    time.sleep(interval / 1000.)
                    steps -= 1

                comm.send({'command': 'stop'})

                data_callback(terminate=True)

            def kill():
                ''' Indicate that the image drawing loop should be killed '''
                state['kill'] = True

            weakref.ref(self, kill)

            threading.Thread(target=do_plot, name=plot_id,
                             args=(self.data_callback,
                                   self.steps, self.interval,
                                   self.annotations, self.transformers)).start()

        try:
            get_ipython().kernel.comm_manager.register_target(plot_id, target_func)
        except NameError:
            warnings.warn('Streaming images are only supported in Jupyter notebooks',
                          RuntimeWarning) 

        styles = []
        styles.append('width: %spx' % (plot_width or self.plot_width))
        styles.append('height: %spx' % (plot_height or self.plot_height))
        styles.append('text-align: right')
        styles.append('background-position: center')
        styles.append('background-repeat: no-repeat')
        if self.size is None:
            pass
        elif isinstance(self.size, six.string_types):
            styles.append('background-size: %s' % self.size)
        elif isinstance(self.size, (int, float)):
            styles.append('background-size: %dpx %dpx' % (self.size, self.size))
        elif isinstance(self.size, (list, tuple)):
            styles.append('background-size: %dpx %dpx' % (self.size[0], self.size[1]))
        styles = '; '.join(styles)

        out = '''
            <div style='position: relative; top: 0; left: 0;
                        height: %(plot_height)spx; width: %(plot_width)spx'>
                <div id='%(plot_id)s_output_bg' style='%(styles)s;
                         position: absolute; top: 0; left: 0'></div>
                <div id='%(plot_id)s_output' style='%(styles)s;
                         position: absolute; top: 0; left: 0'></div>
            </div>

            <style type="text/css">
            #%(plot_id)s_output {
                color: black;
                text-shadow: #fff 0px 0px 1px;
                -webkit-font-smoothing: antialiased;
            }
            .stream-plot-controls {
                border: 1px solid #e0e0e0;
                border-radius: 5px;
                padding: 8px;
                display: inline-block;
            }
            .stream-plot-control {
                color: #909090;
                font-size: 150%%;
                margin-right: 1em;
                margin-left: 1em;
            }
            .fa-play.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #00c000;
                color: #00c000;
            }
            .fa-pause.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #fdeb61;
                color: #d37a2d;
            }
            .fa-stop.stream-plot-control:not([class*='enabled']):not([class*='disabled']):hover {
                text-shadow: 0px 0px 15px #ff0000;
                color: #ff0000;
            }
            .fa-play.stream-plot-control.enabled { color: #00c000; }
            .fa-pause.stream-plot-control.enabled { color: #d37a2d; }
            .fa-stop.stream-plot-control.enabled { color: #ff0000; }
            .stream-plot-control.disabled { color: #d0d0d0; }
            .stream-image-label { background-color: rgba(255, 255, 255, 0.5) }
            </style>

            <div id="streaming-plot-%(plot_id)s"
                 style="text-align: center; padding-top:10px; width: %(plot_width)spx">
            <div class="stream-plot-controls" id="stream-plot-%(plot_id)s">
              <i class="fa fa-play enabled stream-plot-control"
                 title="Stream data into figure" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-play"></i>
              <i class="fa fa-pause stream-plot-control"
                 title="Pause data updates" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-pause"></i>
              <i class="fa fa-stop stream-plot-control"
                 title="Shutdown data stream" aria-hidden="true"
                 style="cursor:pointer" id="%(plot_id)s-stop"></i>
            </div>
            </div>

            <script language='javascript'>
            <!--
            require(['jquery'], function($) {
                var %(plot_id)s = Jupyter.notebook.kernel.comm_manager.new_comm('%(plot_id)s')

                %(plot_id)s.on_msg(function(msg) {
                    var data = msg.content.data;
                    if ( data.url ) {
                        if ( !data.url.startsWith('data:') ) {
                            data.url = atob(data.url);
                        }
                        var preload_img = new Image();
                        preload_img.onload = function () {
                            $('#%(plot_id)s_output_bg').css('background-image',
                                                            'url(' + data.url + ')');
                            setTimeout(function () {
                                $('#%(plot_id)s_output').css('background-image',
                                                             'url(' + data.url + ')')
                                                             .html('<span class="stream-image-label">' +
                                                                   data.format +
                                                                   ' (' + data.size[0] + 'px, ' +
                                                                          data.size[1] + 'px)' +
                                                                   '</span>');
                            }, 15);
                        }
                        preload_img.src = data.url;
                    }

                    if ( data.command ) {
                        if ( data.command == 'stop' ) {
                            $('#%(plot_id)s-play, #%(plot_id)s-pause, #%(plot_id)s-stop')
                                .removeClass('enabled')
                                .addClass('disabled')
                                .css('cursor', 'default');
                            $('#streaming-plot-%(plot_id)s').hide().remove();
                        }
                    }

                    // If plot doesn't exist anymore, stop data
                    if ( $('#streaming-plot-%(plot_id)s').length == 0 ) {
                        %(plot_id)s.send({'command': 'stop'});
                    }
                });

                %(plot_id)s.on_close(function() {
                    %(plot_id)s.send({'command': 'stop'});
                });

                // Send message when plot is removed
                $.event.special.destroyed = {
                    remove: function(o) {
                      if (o.handler) {
                        o.handler()
                      }
                    }
                }
                $('#streaming-plot-%(plot_id)s').bind('destroyed', function() {
                    %(plot_id)s.send({'command': 'stop'});
                });

                var blink = function () {
                  $('#%(plot_id)s-pause.enabled').fadeOut(900).fadeIn(900, blink);
                }

                $('#%(plot_id)s-play').click(function() {
                  if ( $('#%(plot_id)s-play').hasClass('disabled') ) return;
                  if ( $('#%(plot_id)s-play').hasClass('enabled') ) return;
                  $('#%(plot_id)s-pause').removeClass('enabled');
                  $('#%(plot_id)s-play').addClass('enabled');
                  %(plot_id)s.send({'command': 'play'});
                });

                $('#%(plot_id)s-pause').click(function() {
                  if ( $('#%(plot_id)s-pause').hasClass('disabled') ) return;
                  if ( $('#%(plot_id)s-pause').hasClass('enabled') ) {
                    $('#%(plot_id)s-play').click();
                    return;
                  }
                  $('#%(plot_id)s-play').removeClass('enabled');
                  $('#%(plot_id)s-pause').addClass('enabled');
                  %(plot_id)s.send({'command': 'pause'});
                  blink();
                });

                $('#%(plot_id)s-stop').click(function() {
                  if ( $('#%(plot_id)s-stop').hasClass('disabled') ) return;
                    $('#%(plot_id)s-play, #%(plot_id)s-pause, #%(plot_id)s-stop')
                    .removeClass('enabled')
                    .addClass('disabled')
                    .css('cursor', 'default');
                  %(plot_id)s.send({'command': 'stop'});
                  $('#streaming-plot-%(plot_id)s').hide().remove();
                });

            });
            //-->
            </script>
        ''' % dict(plot_id=plot_id, styles=styles,
                   plot_width=plot_width or self.plot_width,
                   plot_height=plot_height or self.plot_height)

        return out


class ChartLayout(object):
    '''
    Create a layout of charts / images

    The layout of chart and image components is specified in the positional
    arguments of the constructor.  Each positional argument becomes a row
    in the layout.  If the argument is a tuple, each item in the tuple becomes
    a cell in that row.

    Parameters
    ----------
    *layout : one-or-more tuple or StreamingImage or StreamingChart
        Components of the layout
    width : int, optional
        The pixel width of the layout
    height : int, optional
        The pixel height of the layout

    Examples
    --------
    The following code will create a layout of two line charts on the first
    row, a streaming image in the second row, and two streaming scatter
    plots on the third row.

    >>> layout = ChartLayout(
    ...              (win.streaming_line('x', 'y'),
    ...               win.streaming_line('x', 'z')),
    ...              win.streaming_images('img'),
    ...              (win2.streaming_scatter('a', 'b'),
    ...               win2.streaming_scatter('a', 'c'))
    ...          )
    >>> layout

    Returns
    -------
    :class:`ChartLayout`

    '''

    def __init__(self, *layout, **kwargs):
        layout = list(layout)
        for i, item in enumerate(layout):
            if not isinstance(item, (list, tuple)):
                layout[i] = [layout[i]]
        self.layout = layout
        self.width = int(kwargs.get('width', 900))
        self.height = int(kwargs.get('height', 300))

    def _repr_html_(self):
        layout_id = 'chart_layout_%s' % str(uuid.uuid4()).replace('-', '_')

        out = []

        out.append('<div id="%s">' % layout_id)

        out.append('''
            <style>
            #%s .stream-plot-controls { display: none }
            </style>
        ''' % layout_id)

        for row in self.layout:
            out.append(('<table style="width: %dpx; height: %dpx; '
                        'border-style: none">') % (self.width, self.height))
            out.append('<tr style="border-style: none">')
            for cell in row:
                cellwidth = int(self.width / len(row))
                cellheight = self.height
                out.append('<td style="width:%dpx; border-style: none">%s</td>' %
                           (cellwidth, cell._repr_html_(plot_width=cellwidth,
                                                        plot_height=cellheight)))
            out.append('</tr>')
            out.append('</table>')

        out.append('</div>')

        out.append('''
            <div id="streaming-layout-%(layout_id)s"
                 style="text-align: center; padding-top:10px; width: %(layout_width)spx">
                <div class="stream-plot-controls" id="stream-layout-%(layout_id)s">
                  <i class="fa fa-play enabled stream-plot-control"
                     title="Stream data into figure" aria-hidden="true"
                     style="cursor:pointer" id="%(layout_id)s-play"></i>
                  <i class="fa fa-pause stream-plot-control"
                     title="Pause data updates" aria-hidden="true"
                     style="cursor:pointer" id="%(layout_id)s-pause"></i>
                  <i class="fa fa-stop stream-plot-control"
                     title="Shutdown data stream" aria-hidden="true"
                     style="cursor:pointer" id="%(layout_id)s-stop"></i>
                </div>
            </div>

            <script>
            <!--
            require(['jquery'], function($) {
                var blink = function () {
                  $('#%(layout_id)s-pause.enabled').fadeOut(900).fadeIn(900, blink);
                }

                $('#%(layout_id)s-play').click(function() {
                    if ( $('#%(layout_id)s-play').hasClass('disabled') ) return;
                    if ( $('#%(layout_id)s-play').hasClass('enabled') ) return;
                    $('#%(layout_id)s-pause').removeClass('enabled');
                    $('#%(layout_id)s-play').addClass('enabled');
                    $('#%(layout_id)s .stream-plot-control[id$="-play"]').click();
                });

                $('#%(layout_id)s-pause').click(function() {
                    if ( $('#%(layout_id)s-pause').hasClass('disabled') ) return;
                    if ( $('#%(layout_id)s-pause').hasClass('enabled') ) {
                        $('#%(layout_id)s-play').click();
                        return;
                    }
                    $('#%(layout_id)s-play').removeClass('enabled');
                    $('#%(layout_id)s-pause').addClass('enabled');
                    $('#%(layout_id)s .stream-plot-control[id$="-pause"]').click();
                    blink();
                });

                $('#%(layout_id)s-stop').click(function() {
                    if ( $('#%(layout_id)s-stop').hasClass('disabled') ) return;
                        $('#%(layout_id)s-play, #%(layout_id)s-pause, #%(layout_id)s-stop')
                        .removeClass('enabled')
                        .addClass('disabled')
                        .css('cursor', 'default');
                    $('#%(layout_id)s .stream-plot-control[id$="-stop"]').click();
                    $('#streaming-layout-%(layout_id)s').hide().remove();
                });
            });
            //-->
            </script>
        ''' % dict(layout_id=layout_id, layout_width=self.width, layout_height=self.height))

        return '\n'.join(out)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Metadata '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
from six.moves import urllib
from .base import RESTHelpers


class Metadata(collections.MutableMapping, RESTHelpers):
    '''
    Base class for metadata dictionaries

    This is an abstract class that must be subclassed.  You must
    implement the ``__metadata__``, ``__setitem__``, and ``__delitem__``
    methods.

    Parameters
    ----------
    session : requests.Session
        The session object

    '''

    def __init__(self, session=None):
        RESTHelpers.__init__(self, session=session)

    def __metadata__(self):
        raise NotImplementedError

    def _set_metadata_from_xml(self, item, out):
        out[item.attrib['id']] = item.text

    def __getitem__(self, key):
        return self.__metadata__()[key]

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        return iter(self.__metadata__())

    def __len__(self):
        return len(self.__metadata__())

    def __str__(self):
        return str(self.__metadata__())

    def __repr__(self):
        return repr(self.__metadata__())
